"""
=====
Steps
=====

This module is responsible for defining the abstractions that represent desired
steps to run in a pipeline. These so-called "steps" are high level and do not indicate
how they are to actually be implemented.

"""

from __future__ import annotations

import copy
from abc import ABC, abstractmethod
from collections import defaultdict
from collections.abc import Iterable
from typing import Any

from layered_config_tree import LayeredConfigTree

from easylink.graph_components import (
    EdgeParams,
    ImplementationGraph,
    InputSlot,
    InputSlotMapping,
    OutputSlot,
    OutputSlotMapping,
    SlotMapping,
    StepGraph,
)
from easylink.implementation import (
    Implementation,
    NullImplementation,
    PartialImplementation,
)
from easylink.utilities import paths
from easylink.utilities.data_utils import load_yaml

COMBINED_IMPLEMENTATION_KEY = "combined_implementation_key"


class Step:
    """The highest-level pipeline building block abstraction.

    ``Steps`` contain information about the purpose of the interoperable tasks in
    the sequence called a "pipeline" and how those tasks relate to one another.
    In turn, ``Steps`` are implemented by :class:`Implementations<easylink.implementation.Implementation>`,
    such that each ``Step`` may have several ``Implementations`` to choose from
    but each ``Implementation`` must implemement exactly one ``Step``. As such, the
    pipeline for a given EasyLink run consists of ``Implementations`` that collectively
    span the ``Steps`` in the :class:`~easylink.pipeline_schema.PipelineSchema`.

    Parameters
    ----------
    step_name
        The name of the pipeline step in the ``PipelineSchema``.
    name
        The name of this step *node*. This can be different from the ``step_name``
        due to the need for disambiguation during the process of unrolling loops,
        etc. For example, if step 1 is looped multiple times, each node would
        have a ``step_name`` of, perhaps, "step_1" but unique ``names``
        ("step_1_loop_1", etc).
    input_slots
        All required :class:`InputSlots<easylink.graph_components.InputSlot>`.
    output_slots
        All required :class:`OutputSlots<easylink.graph_components.OutputSlot>`.
    nodes
        All sub-nodes (i.e. sub-``Steps``) of this particular ``Step`` instance.
    edges
        The :class:`~easylink.graph_components.EdgeParams` of this ``Step``.
    input_slot_mappings
        The :class:`InputSlotMapping<easylink.graph_components.InputSlotMapping>` of this ``Step``.
    output_slot_mappings
        The :class:`OutputSlotMapping<easylink.graph_components.OutputSlotMapping>` of this ``Step``.

    Notes
    -----
    This is the most basic type of step object available in the pipeline; it
    represents a single element of work to be run one time in the pipeline. Other
    classes inherit from this and expand upon it to represent more complex structures,
    e.g. to loop a step multiple times or to run multiple steps in parallel.

    """

    def __init__(
        self,
        step_name: str,
        name: str | None = None,
        input_slots: Iterable[InputSlot] = (),
        output_slots: Iterable[OutputSlot] = (),
        nodes: Iterable[Step] = (),
        edges: Iterable[EdgeParams] = (),
        input_slot_mappings: Iterable[InputSlotMapping] = (),
        output_slot_mappings: Iterable[OutputSlotMapping] = (),
    ) -> None:
        self.step_name = step_name
        """The name of the high-level pipeline step."""
        self.name = name if name else step_name
        """The name of ``Step's`` node in its :class:`~easylink.graph_components.StepGraph`. 
        This is a more descriptive name than the ``step_name``, e.g. if "step 1" 
        is looped multiple times. If not provided, defaults to the :attr:`step_name`."""
        self.input_slots = {slot.name: slot for slot in input_slots}
        """A mapping of ``InputSlot`` names to their instances."""
        self.output_slots = {slot.name: slot for slot in output_slots}
        """A mapping of ``OutputSlot`` names to their instances."""
        self.nodes = nodes
        """All sub-nodes (i.e. sub-``Steps``) of this particular ``Step`` instance."""
        for node in self.nodes:
            node.set_parent_step(self)
        self.edges = edges
        """The :class:`~easylink.graph_components.EdgeParams` of this ``Step``."""
        self.step_graph = self._get_step_graph(nodes, edges)
        """The :class:`~easylink.graph_components.StepGraph` of this ``Step``, i.e.
        the directed acyclic graph (DAG) of sub-nodes and their edges that make 
        up this ``Step`` instance."""
        self.slot_mappings = {
            "input": list(input_slot_mappings),
            "output": list(output_slot_mappings),
        }
        """A combined dictionary containing both the ``InputSlotMappings`` and
        ``OutputSlotMappings`` of this ``Step``."""
        self.parent_step = None
        """This ``Step's`` parent ``Step``, if applicable."""
        self._configuration_state = None
        """This ``Step's`` :class:`~easylink.step.ConfigurationState`."""

    @property
    def config_key(self):
        """The configuration key pertinent to this type of ``Step``."""
        return None

    @property
    def configuration_state(self) -> ConfigurationState:
        """The :class:`~easylink.step.ConfigurationState` of this ``Step``."""
        if self._configuration_state is None:
            raise ValueError(
                f"Step {self.name}'s configuration_state was invoked before being set"
            )
        return self._configuration_state

    @property
    def implementation_node_name(self) -> str:
        """The unique name to be used for this ``Step's`` node in the :class:`~easylink.graph_components.ImplementationGraph`.

        This compares the ``Step`` *instance* name to its *node* name via the ``Step's``
        ordered hierarchy of sub-``Steps`` and uses the full suffix of names starting
        from wherever the two first differ.

        For example, a ``Step`` named "step_3" may loop multiple times using the same
        :class:`~easylink.implementation.Implementation` named "step_3_python_pandas".
        However, to disambiguate between the different loops of "step_3", we might
        designate the node name to be "step_3_loop_1" and then combine that with the
        ``Implementation`` name such that the ``Implementation's`` node name is
        "step_3_loop_1_step_3_python_pandas".

        If all the node names and step names match, we have not introduced any step
        degeneracies (with e.g. loops or multiples), and we can simply use the
        implementation name directly.

        Returns
        -------
            The unique name to be used for this ``Step's`` node in the ``ImplementationGraph``.
        """
        step = self
        implementation_name = (
            self.configuration_state.pipeline_config[COMBINED_IMPLEMENTATION_KEY]
            if self.configuration_state.is_combined
            else self.configuration_state.implementation_config.name
        )
        node_names = []
        step_names = []
        while step:
            node_names.append(step.name)
            step_names.append(step.step_name)
            step = step.parent_step

        prefix = []
        step_names.reverse()
        node_names.reverse()
        for i, (step_name, node_name) in enumerate(zip(step_names, node_names)):
            if step_name != node_name:
                prefix = node_names[i:]
                break
        # If we didn't include the step name already for a combined implementation, do so now.
        if self.configuration_state.is_combined and not prefix:
            prefix.append(self.name)
        prefix.append(implementation_name)
        return "_".join(prefix)

    ###########
    # Methods #
    ###########

    def validate_step(
        self,
        step_config: LayeredConfigTree,
        combined_implementations: LayeredConfigTree,
        input_data_config: LayeredConfigTree,
    ) -> dict[str, list[str]]:
        """Validates the ``Step``.

        Parameters
        ----------
        step_config
            The configuration of this ``Step``.
        combined_implementations
            The configuration for any implementations to be combined.
        input_data_config
            The input data configuration for the entire pipeline.

        Returns
        -------
            A dictionary of errors, where the keys are the ``Step`` name and the
            values are lists of error messages associated with the given ``Step``.

        Notes
        -----
        A ``Step`` can be in either a "leaf" or a "non-leaf" configuration state
        and the validation process is different for each.

        If the ``Step`` does not validate (i.e. errors are found and the returned
        dictionary is non-empty), the tool will exit and the pipeline will not run.

        We attempt to batch error messages as much as possible, but there may be
        times where the configuration is so ill-formed that we are unable to handle
        all issues in one pass. In these cases, new errors may be found after the
        initial ones are handled.
        """
        if len(self.step_graph.nodes) == 0:
            return self._validate_leaf(step_config, combined_implementations)
        elif self.config_key in step_config:
            return self._validate_nonleaf(
                step_config[self.config_key], combined_implementations, input_data_config
            )
        else:
            return self._validate_leaf(step_config, combined_implementations)

    def get_implementation_graph(self) -> ImplementationGraph:
        """Gets this ``Step's`` :class:`~easylink.graph_components.ImplementationGraph`.

        The ``ImplementationGraph`` and how it is determined depends on whether
        this ``Step`` is a leaf or a non-leaf, i.e. what its :attr:`configuration_state`
        is.

        Returns
        -------
            The ``ImplementationGraph`` of this ``Step`` based on its ``configuration_state``.
        """
        return self.configuration_state.get_implementation_graph()

    def get_implementation_edges(self, edge: EdgeParams) -> list[EdgeParams]:
        """Gets the edge information for the ``Implementation`` related to this ``Step``.

        Parameters
        ----------
        edge
            The ``Step's`` edge information to be propagated to the ``ImplementationGraph``.

        Returns
        -------
            The ``Implementation's`` edge information based on this ``Step's`` configuration
            state.
        """
        return self.configuration_state.get_implementation_edges(edge)

    def set_parent_step(self, step: Step) -> None:
        """Sets the parent of this ``Step``.

        Parameters
        ----------
        step
            The parent ``Step`` to be set for this instance's :attr:`parent_step`.
        """
        self.parent_step = step

    def set_configuration_state(
        self,
        parent_config: LayeredConfigTree,
        combined_implementations: LayeredConfigTree,
        input_data_config: LayeredConfigTree,
    ) -> None:
        """Sets the configuration state for this ``Step``.

        The so-called 'configuration state' for a given ``Step`` is backed up by
        a :class:`ConfigurationState` class and is assigned to its :attr:`_configuration_state`
        attribute. There are two possible ``ConfigurationStates``:
        :class:`LeafConfigurationState` and :class:`NonLeafConfigurationState`.

        This method sets the configuration state of this ``Step`` based on whether
        or not a :attr:`config_key` is set *and exists is the ``Step's`` configuration*
        (i.e. its portion of the user-suppled pipeline specification
        file); any required deviation from this behavior requires special
        handling.

        Parameters
        ----------
        parent_config
            The configuration of the parent ``Step``.
        combined_implementations
            The configuration for any implementations to be combined.
        input_data_config
            The input data configuration for the entire pipeline.
        """
        step_config = parent_config[self.name]
        sub_config = self._get_config(step_config)
        if self.config_key is not None and self.config_key in step_config:
            self._configuration_state = NonLeafConfigurationState(
                self, sub_config, combined_implementations, input_data_config
            )
        else:
            self._configuration_state = LeafConfigurationState(
                self, sub_config, combined_implementations, input_data_config
            )

    def get_implementation_slot_mappings(self) -> dict[str, list[SlotMapping]]:
        """Gets the input and output :class:`SlotMappings<easylink.graph_components.SlotMapping>`."""
        return {
            "input": [
                InputSlotMapping(slot, self.implementation_node_name, slot)
                for slot in self.input_slots
            ],
            "output": [
                OutputSlotMapping(slot, self.implementation_node_name, slot)
                for slot in self.output_slots
            ],
        }

    ##################
    # Helper methods #
    ##################

    def _get_step_graph(self, nodes: list[Step], edges: list[EdgeParams]) -> StepGraph:
        """Create a StepGraph from the nodes and edges the step was initialized with."""
        step_graph = StepGraph()
        for step in nodes:
            step_graph.add_node_from_step(step)
        for edge in edges:
            step_graph.add_edge_from_params(edge)
        return step_graph

    def _validate_leaf(
        self,
        step_config: LayeredConfigTree,
        combined_implementations: LayeredConfigTree,
    ) -> dict[str, list[str]]:
        """Validates a leaf ``Step``."""
        errors = {}
        metadata = load_yaml(paths.IMPLEMENTATION_METADATA)
        error_key = f"step {self.name}"
        if (
            "implementation" not in step_config
            and COMBINED_IMPLEMENTATION_KEY not in step_config
        ):
            errors[error_key] = [
                "The step configuration does not contain an 'implementation' key or a "
                "reference to a combined implementation."
            ]
        elif (
            COMBINED_IMPLEMENTATION_KEY in step_config
            and not step_config[COMBINED_IMPLEMENTATION_KEY] in combined_implementations
        ):
            errors[error_key] = [
                f"The step refers to a combined implementation but {step_config[COMBINED_IMPLEMENTATION_KEY]} is not a "
                f"valid combined implementation."
            ]
        else:
            implementation_config = (
                step_config["implementation"]
                if "implementation" in step_config
                else combined_implementations[step_config[COMBINED_IMPLEMENTATION_KEY]]
            )
            if not "name" in implementation_config:
                errors[error_key] = [
                    "The implementation configuration does not contain a 'name' key."
                ]
            elif not implementation_config["name"] in metadata:
                errors[error_key] = [
                    f"Implementation '{implementation_config['name']}' is not supported. "
                    f"Supported implementations are: {list(metadata.keys())}."
                ]
        return errors

    def _validate_nonleaf(
        self,
        step_config: LayeredConfigTree,
        combined_implementations: LayeredConfigTree,
        input_data_config: LayeredConfigTree,
    ) -> dict[str, list[str]]:
        """Validates a non-leaf ``Step``."""
        errors = {}
        nodes = self.step_graph.nodes
        for node in nodes:
            step = nodes[node]["step"]
            if isinstance(step, IOStep):
                continue
            if step.name not in step_config:
                step_errors = {f"step {step.name}": [f"The step is not configured."]}
            else:
                step_errors = step.validate_step(
                    step_config[step.name], combined_implementations, input_data_config
                )
            if step_errors:
                errors.update(step_errors)
        extra_steps = set(step_config.keys()) - set(nodes)
        for extra_step in extra_steps:
            errors[f"step {extra_step}"] = [f"{extra_step} is not a valid step."]
        return errors

    def _get_config(self, step_config: LayeredConfigTree) -> LayeredConfigTree:
        """Convenience method to get a ``Step's`` configuration.

        Some types of ``Steps`` have a unique :attr:`config_key` (defined by the
        user via the pipeline specification file) that is used to specify the behavior
        of the ``Step`` (e.g. looping, parallel, etc). This method simply returns
        the ``Step's`` sub-configuration keyed to that ``config_key`` (if it exists,
        i.e. is not a basic ``Step``).

        Parameters
        ----------
        step_config
            The high-level configuration of this ``Step``.

        Returns
        -------
            The sub-configuration of this ``Step`` keyed on the ``config_key``
            (if it exists).

        """
        return (
            step_config
            if not self.config_key in step_config
            else step_config[self.config_key]
        )


class IOStep(Step):
    """A special case type of :class:`Step` used to represent incoming and outgoing data.

    ``IOSteps`` are used to handle the incoming and outgoing data to the pipeline;
    they are inherited by concrete :class:`InputStep` and :class:`OutputStep`
    classes. These are not typical ``Steps`` in that they do not represent a unit
    of work to be performed in the pipeline (i.e. there is no container to run) and,
    thus, are not implemented by an :class:`~easylink.implementation.Implementation`.

    See :class:`Step` for inherited attributes.

    """

    @property
    def implementation_node_name(self) -> str:
        """Dummy name to allow ``IOSteps`` to be used interchangeably with other ``Steps``.

        Unlike other types of ``Steps``, ``IOSteps`` are not actually implemented
        via an :class:`~easylink.implementation.Implementation` and thus do not
        require a different node name than its own ``Step`` name. This property
        only exists so that ``IOSteps`` can be used interchangeably with other
        ``Steps`` in the codebase.

        Returns
        -------
            The ``IOStep's`` name.
        """
        return self.name

    def validate_step(
        self,
        step_config: LayeredConfigTree,
        combined_implementations: LayeredConfigTree,
        input_data_config: LayeredConfigTree,
    ) -> dict[str, list[str]]:
        """Dummy validation method to allow ``IOSteps`` to be used interchangeably with other ``Steps``.

        Unlike other types of ``Steps``, ``IOSteps`` are not actually implemented
        via an :class:`~easylink.implementation.Implementation` and thus do not
        require any sort of validation since no new data is created. This method
        only exists so that ``IOSteps`` can be used interchangeably with other
        ``Steps`` in the codebase.

        Returns
        -------
            An empty dictionary.
        """
        return {}

    def set_configuration_state(
        self,
        parent_config: LayeredConfigTree,
        combined_implementations: LayeredConfigTree,
        input_data_config: LayeredConfigTree,
    ) -> None:
        """Sets the configuration state to leaf.

        An ``IOStep`` is by definition a leaf ``Step`` and so we assign that here
        instead of relying on the default behavior of the parent class.

        Parameters
        ----------
        parent_config
            The configuration of the parent ``Step``. For ``IOSteps``, this will
            always be the entire pipeline configuration.
        combined_implementations
            The configuration for any implementations to be combined.
        input_data_config
            The input data configuration for the entire pipeline.
        """
        self._configuration_state = LeafConfigurationState(
            self, parent_config, combined_implementations, input_data_config
        )

    def get_implementation_graph(self) -> ImplementationGraph:
        """Gets this ``Step's`` :class:`~easylink.graph_components.ImplementationGraph`.

        Notes
        -----
        Unlike other types of ``Steps``, ``IOSteps`` are not actually implemented
        via an :class:`~easylink.implementation.Implementation`. As such, we
        leverage the :class:`~easylink.implementation.NullImplementation` class
        to generate the graph node.

        Returns
        -------
            The ``ImplementationGraph`` of this ``Step``.
        """
        implementation_graph = ImplementationGraph()
        implementation_graph.add_node_from_implementation(
            self.name,
            implementation=NullImplementation(
                self.name, self.input_slots.values(), self.output_slots.values()
            ),
        )
        return implementation_graph


class InputStep(IOStep):
    """A special case type of :class:`Step` used to represent incoming data.

    An ``InputStep`` is used to pass data into the pipeline. Since we do not know
    what the data to pass into the pipeline will be a priori, we instantiate an
    "all" :class:`~easylink.graph_components.OutputSlot` which is used to pass in
    *all* data defined in the input data specification file.

    See :class:`IOStep` for inherited attributes.
    """

    def __init__(self) -> None:
        super().__init__(step_name="input_data", output_slots=(OutputSlot("all"),))

    def set_configuration_state(
        self,
        parent_config: LayeredConfigTree,
        combined_implementations: LayeredConfigTree,
        input_data_config: LayeredConfigTree,
    ) -> None:
        """Sets the configuration state and updates the ``OutputSlots``.

        In addition to setting ``InputStep`` to a leaf configuration state, this
        method also updates the ``OutputSlots`` to include all of the dataset keys
        in the input data specification file. This allows for future use of
        specific datasets instead of only "all" of them.

        Parameters
        ----------
        parent_config
            The configuration of the parent ``Step``. For ``IOSteps``, this will
            always be the entire pipeline configuration.
        combined_implementations
            The configuration for any implementations to be combined.
        input_data_config
            The input data configuration for the entire pipeline.
        """
        super().set_configuration_state(
            parent_config, combined_implementations, input_data_config
        )
        for input_data_key in input_data_config:
            self.output_slots[input_data_key] = OutputSlot(name=input_data_key)


class OutputStep(IOStep):
    """A special case type of :class:`Step` used to represent final results data.

    An ``OutputStep`` is used to write the `Snakemake <https://snakemake.readthedocs.io/en/stable/>`_
    Snakefile target rule in the :meth:`easylink.pipeline.Pipeline.build_snakefile`
    method.

    See :class:`IOStep` for inherited attributes.

    """

    def __init__(self, input_slots: Iterable[InputSlot]) -> None:
        super().__init__("results", input_slots=input_slots)


class HierarchicalStep(Step):
    """A type of :class:`Step` that can may contain sub-``Steps``.

    A ``HierarchicalStep`` can be represented by multiple sub-``Steps`` (and thus
    implemented by the sub-``Steps'`` respective :class:`Implementations<easylink.implementation.Implementation>`.
    For example, "step_1" might be represented by a "step_1a" and a "step_1b", each
    of which has its own ``Implementation``.

    See :class:`Step` for inherited attributes.

    Notes
    -----
    To use this feature, the sub-``Steps`` must be defined in the pipeline specification
    file under a "substeps" key. If no "substeps" key is present, it will be  treated
    as a single ``Step``.

    """

    @property
    def config_key(self):
        """The pipeline specification key required for a ``HierarchicalStep``."""
        return "substeps"


class TemplatedStep(Step, ABC):
    """A type of :class:`Step` that may contain multiplicity.

    A ``TemplatedStep`` is used to represents a ``Step`` that contains a specified
    amount of multiplicity, such as one that is looped or run in parallel; it is
    inherited by concrete :class:`LoopStep` and :class:`ParallelStep` instances.

    See :class:`Step` for inherited attributes.

    Parameters
    ----------
    template_step
        The ``Step`` to be templated.

    """

    def __init__(
        self,
        template_step: Step,
    ) -> None:
        super().__init__(
            template_step.step_name,
            template_step.name,
            template_step.input_slots.values(),
            template_step.output_slots.values(),
        )
        self.template_step = template_step
        """The ``Step`` to be templated."""
        self.template_step.set_parent_step(self)

    @property
    @abstractmethod
    def node_prefix(self) -> str:
        """The prefix to be used in the node name.

        To disambiguate between the different types of nodes with multiplicity
        (i.e. loops or parallel), we use a unique prefix to be used as necessary.

        Returns
        -------
            The prefix to be used for the concrete ``TemplatedStep`` instances.
        """
        pass

    @abstractmethod
    def _update_step_graph(self, num_repeats: int) -> StepGraph:
        """Updates the :class:`~easylink.graph_components.StepGraph`.

        The ``TemplatedStep`` concrete instances must handle the fact that there
        is multiplicity in the ``StepGraph`` and update it accordingly.

        Parameters
        ----------
        num_repeats
            The number of copies to be made of the ``TemplatedStep``.

        Returns
        -------
            The updated ``StepGraph`` with unrolled ``Steps``.

        Notes
        -----
        We do not know a priori - or even during instantiation of the
        :class:`~easylink.pipeline_schema.PipelineSchema` - how many copies of any
        ``TemplatedSteps`` to make; indeed, there may be no ``TemplatedSteps`` at
        all. The user-provided pipeline configuration file must be read in in order
        to determine the number of multiples to generate.
        """
        pass

    @abstractmethod
    def _update_slot_mappings(self, num_repeats: int) -> dict[str, list[SlotMapping]]:
        """Updates the :class:`SlotMappings<easylink.graph_components.SlotMapping>`.

        Parameters
        ----------
        num_repeats
            The number of copies to be made of the ``TemplatedStep``.

        Returns
        -------
            Updated ``SlotMappings`` that account for the ``TemplatedStep`` multiplicity.
        """
        pass

    def validate_step(
        self,
        step_config: LayeredConfigTree,
        combined_implementations: LayeredConfigTree,
        input_data_config: LayeredConfigTree,
    ) -> dict[str, list[str]]:
        """Validates the ``TemplatedStep``.

        Regardless of whether or not a :attr:`Step.config_key` is set, we always
        validate the the base ``Step`` used to create the ``TemplatedStep``. If a
        ``config_key`` is indeed set (that is, there is some multiplicity), we
        complete additional validations.

        Parameters
        ----------
        step_config
            The configuration of this ``TemplatedStep``.
        combined_implementations
            The configuration for any implementations to be combined.
        input_data_config
            The input data configuration for the entire pipeline.

        Returns
        -------
            A dictionary of errors, where the keys are the ``TemplatedStep`` name
            and the values are lists of error messages associated with the given
            ``TemplatedStep``.

        Notes
        -----
        If the ``Step`` does not validate (i.e. errors are found and the returned
        dictionary is non-empty), the tool will exit and the pipeline will not run.

        We attempt to batch error messages as much as possible, but there may be
        times where the configuration is so ill-formed that we are unable to handle
        all issues in one pass. In these cases, new errors may be found after the
        initial ones are handled.
        """
        if not self.config_key in step_config:
            return self.template_step.validate_step(
                step_config, combined_implementations, input_data_config
            )

        sub_config = step_config[self.config_key]

        if not isinstance(sub_config, list):
            return {
                f"step {self.name}": [
                    f"{self.node_prefix.capitalize()} instances must be formatted "
                    "as a sequence in the pipeline configuration."
                ]
            }

        if len(sub_config) == 0:
            return {
                f"step {self.name}": [
                    f"No {self.node_prefix} instances configured under '{self.config_key}' key."
                ]
            }

        errors = defaultdict(dict)
        for i, parallel_config in enumerate(sub_config):
            parallel_errors = {}
            input_data_file = parallel_config.get("input_data_file")
            if input_data_file and not input_data_file in input_data_config:
                parallel_errors["Input Data Key"] = [
                    f"Input data file '{input_data_file}' not found in input data configuration."
                ]
            parallel_errors.update(
                self.template_step.validate_step(
                    parallel_config, combined_implementations, input_data_config
                )
            )
            if parallel_errors:
                errors[f"step {self.name}"][f"{self.node_prefix}_{i+1}"] = parallel_errors
        return errors

    def _get_config(self, step_config: LayeredConfigTree) -> LayeredConfigTree:
        """Convenience method to get the ``TemplatedStep's`` configuration.

        ``TemplatedSteps`` may include multiplicity. In such cases, their configurations
        must be modified to include the expanded ``Steps``.

        Parameters
        ----------
        step_config
            The high-level configuration of this ``TemplatedStep``.

        Returns
        -------
            The expanded sub-configuration of this ``TemplatedStep`` based on the
            :attr:`Step.config_key` and expanded to include all looped or parallelized
            sub-``Steps``).
        """
        if self.config_key in step_config:
            expanded_step_config = LayeredConfigTree()
            for i, sub_config in enumerate(step_config[self.config_key]):
                expanded_step_config.update(
                    {f"{self.name}_{self.node_prefix}_{i+1}": sub_config}
                )
            return expanded_step_config
        return step_config

    def set_configuration_state(
        self,
        parent_config: LayeredConfigTree,
        combined_implementations: LayeredConfigTree,
        input_data_config: LayeredConfigTree,
    ):
        """Sets the configuration state and updates the :class:`SlotMappings<easylink.graph_components.SlotMapping>`.

        Parameters
        ----------
        parent_config
            The configuration of the parent ``Step``.
        combined_implementations
            The configuration for any implementations to be combined.
        input_data_config
            The input data configuration for the entire pipeline.

        Notes
        -----
        A ``TemplatedStep`` is *always* assigned a :class:`NonLeafConfigurationState`
        even if it has no multiplicity since (despite having no copies to make) we
        still need to traverse the sub-``Steps`` to get to the one with a single
        :class:`~easylink.implementation.Implementation`, i.e. the one with a
        :class:`LeafConfigurationState`.
        """
        step_config = parent_config[self.name]
        if self.config_key not in step_config:
            # Special handle the step_graph update
            self.step_graph = StepGraph()
            self.template_step.name = self.name
            self.step_graph.add_node_from_step(self.template_step)
            # Special handle the slot_mappings update
            input_mappings = [
                InputSlotMapping(slot, self.name, slot) for slot in self.input_slots
            ]
            output_mappings = [
                OutputSlotMapping(slot, self.name, slot) for slot in self.output_slots
            ]
            self.slot_mappings = {"input": input_mappings, "output": output_mappings}
            # Add the key back to the expanded config
            expanded_config = LayeredConfigTree({self.name: step_config})
        else:
            expanded_config = self._get_config(step_config)
            num_repeats = len(expanded_config)
            self.step_graph = self._update_step_graph(num_repeats)
            self.slot_mappings = self._update_slot_mappings(num_repeats)
        # Manually set the configuration state to non-leaf instead of relying
        # on super().get_configuration_state() because that method will erroneously
        # set to leaf state in the event the user didn't include the config_key
        # in the pipeline specification.
        self._configuration_state = NonLeafConfigurationState(
            self, expanded_config, combined_implementations, input_data_config
        )

    def _duplicate_template_step(self) -> Step:
        """Makes a duplicate of the template ``Step``.

        Returns
        -------
            A duplicate of the :attr:`template_step`.

        Notes
        -----
        A naive deepcopy would also make a copy of the :attr:`Step.parent_step`; we don't
        want this to be pointing to a *copy* of `self`, but rather to the original.
        We thus re-set the :attr:`Step.parent_step` to the original (`self`) after making
        the copy.
        """
        step_copy = copy.deepcopy(self.template_step)
        step_copy.set_parent_step(self)
        return step_copy


class LoopStep(TemplatedStep):
    """A type of :class:`TemplatedStep` that allows for looping.

    A ``LoopStep`` allows a user to loop a single :class:`Step` or a sequence
    of ``Steps`` multiple times such that each iteration depends on the previous.

    See :class:``TemplatedStep`` for inherited attributes.

    Parameters
    ----------
    template_step
        The ``Step`` to be templated.
    self_edges
        :class:`~easylink.graph_components.EdgeParams` that represent self-edges,
        i.e. edges that connect the output of one loop to the input of the next.

    """

    def __init__(
        self,
        template_step: Step | None = None,
        self_edges: Iterable[EdgeParams] = (),
    ) -> None:
        super().__init__(template_step)
        self.self_edges = self_edges
        """:class:`~easylink.graph_components.EdgeParams` that represent self-edges,
        i.e. edges that connect the output of one loop to the input of the next."""

    @property
    def config_key(self):
        """The pipeline specification key required for a ``LoopStep``."""
        return "iterate"

    @property
    def node_prefix(self):
        """The prefix to be used in the ``LoopStep`` node name."""
        return "loop"

    def _update_step_graph(self, num_repeats) -> StepGraph:
        """Updates the :class:`~easylink.graph_components.StepGraph` to include loops.

        This makes ``num_repeats`` copies of the :class:`TemplatedStep` and chains
        them together sequentially according to the self edges.

        Parameters
        ----------
        num_repeats
            The number of loops.

        Returns
        -------
            The updated ``StepGraph`` with ``num_repeats`` serial :class:`Steps<Step>`
            and their corrected edges.
        """
        graph = StepGraph()
        nodes = []
        edges = []

        for i in range(num_repeats):
            updated_step = self._duplicate_template_step()
            updated_step.name = f"{self.name}_{self.node_prefix}_{i+1}"
            nodes.append(updated_step)
            if i > 0:
                for self_edge in self.self_edges:
                    source_node = f"{self.name}_{self.node_prefix}_{i}"
                    target_node = f"{self.name}_{self.node_prefix}_{i+1}"
                    edge = EdgeParams(
                        source_node=source_node,
                        target_node=target_node,
                        input_slot=self_edge.input_slot,
                        output_slot=self_edge.output_slot,
                    )
                    edges.append(edge)

        for node in nodes:
            graph.add_node_from_step(node)
        for edge in edges:
            graph.add_edge_from_params(edge)
        return graph

    def _update_slot_mappings(self, num_repeats) -> dict[str, list[SlotMapping]]:
        """Updates the :class:`SlotMappings<easylink.graph_components.SlotMapping>`.

        This updates the appropriate slot mappings based on the number of loops
        and the non-self-edge input and output slots.

        Parameters
        ----------
        num_repeats
            The number of loops.

        Returns
        -------
            Updated ``SlotMappings`` that account for the number of loops requested.
        """
        input_mappings = []
        self_edge_input_slots = {edge.input_slot for edge in self.self_edges}
        external_input_slots = self.input_slots.keys() - self_edge_input_slots
        for input_slot in self_edge_input_slots:
            input_mappings.append(
                InputSlotMapping(input_slot, f"{self.name}_{self.node_prefix}_1", input_slot)
            )
        for input_slot in external_input_slots:
            input_mappings.extend(
                [
                    InputSlotMapping(
                        input_slot, f"{self.name}_{self.node_prefix}_{n+1}", input_slot
                    )
                    for n in range(num_repeats)
                ]
            )
        output_mappings = [
            OutputSlotMapping(slot, f"{self.name}_{self.node_prefix}_{num_repeats}", slot)
            for slot in self.output_slots
        ]
        return {"input": input_mappings, "output": output_mappings}


class ParallelStep(TemplatedStep):
    """A type of :class:`TemplatedStep` that creates multiple copies in parallel
    with no dependencies between them.

    See :class:`TemplatedStep` for inherited attributes.

    """

    @property
    def config_key(self):
        """The pipeline specification key required for a ``ParallelStep``."""
        return "parallel"

    @property
    def node_prefix(self):
        """The prefix to be used in the ``ParallelStep`` node name."""
        return "parallel_split"

    def _update_step_graph(self, num_repeats: int) -> StepGraph:
        """Updates the :class:`~easylink.graph_components.StepGraph` to include parallelization.

        This makes ``num_repeats`` copies of the ``TemplatedStep`` that are
        independent but contain the same edges.

        Parameters
        ----------
        num_repeats
            The number of parallel ``TemplatedSteps``.

        Returns
        -------
            The updated ``StepGraph`` with ``num_repeats`` parallel :class:`Steps<Step>`
            and their corrected edges.
        """
        graph = StepGraph()

        for i in range(num_repeats):
            updated_step = self._duplicate_template_step()
            updated_step.name = f"{self.name}_{self.node_prefix}_{i+1}"
            graph.add_node_from_step(updated_step)
        return graph

    def _update_slot_mappings(self, num_repeats: int) -> dict[str, list[SlotMapping]]:
        """Updates the :class:`SlotMappings<easylink.graph_components.SlotMapping>`.

        This updates the appropriate slot mappings based on the number of parallel
        copies and the existing input and output slots.

        Parameters
        ----------
        num_repeats
            The number of parallel copies.

        Returns
        -------
            Updated ``SlotMappings`` that account for the number of copies requested.
        """
        input_mappings = [
            InputSlotMapping(slot, f"{self.name}_{self.node_prefix}_{n+1}", slot)
            for n in range(num_repeats)
            for slot in self.input_slots
        ]
        output_mappings = [
            OutputSlotMapping(slot, f"{self.name}_{self.node_prefix}_{n+1}", slot)
            for n in range(num_repeats)
            for slot in self.output_slots
        ]
        return {"input": input_mappings, "output": output_mappings}


class EmbarrassinglyParallelStep(Step):
    """A step that is run in parallel on the backend.

    An ``EmbarrassinglyParallelStep`` is different than a :class:`ParallelStep`
    in that it is not configured by the user to be run in parallel - it completely
    happens on the back end for performance reasons. As such, note that it inherits
    from :class:`Step` instead of :class:`TemplatedStep`.
    """

    def __init__(
        self,
        step_name: str,
        input_slots: Iterable[InputSlot],
        output_slots: Iterable[OutputSlot],
    ) -> None:
        super().__init__(step_name, input_slots=input_slots, output_slots=output_slots)
        self._validate()

    def _validate(self) -> None:
        """Validates the ``EmbarrassinglyParallelStep``.

        ``EmbarrassinglyParallelSteps`` are not configured by the user to be run
        in parallel. Since it happens on the back end, we need to do somewhat unique
        validations during construction. Specifically,
        - one and only one :class:`~easylink.graph_components.InputSlot` *must* include
        a :attr:`~easylink.graph_components.InputSlot.splitter` method.
        - all :class:`OutputSlots<easylink.graph_components.OutputSlot>` *must* include
        an :attr:`~easylink.graph_components.OutputSlot.aggregator` method.
        """
        errors = []
        # assert that only one input slot has a splitter assigned
        splitters = {
            slot.name: slot.splitter.__name__
            for slot in self.input_slots.values()
            if slot.splitter
        }
        if len(splitters) == 0:
            errors.append(
                f"EmbarrassinglyParallelStep '{self.step_name}' does not have any input slots with a "
                "splitter method assigned; one and only one input slot must have a splitter."
            )
        if len(splitters) > 1:
            errors.append(
                f"EmbarrassinglyParallelStep '{self.step_name}' has multiple input slots with "
                "splitter methods assigned; one and only one input slot must have a splitter.\n"
                f"Input slots with splitters: {splitters}"
            )
        missing_aggregators = [
            slot.name for slot in self.output_slots.values() if not slot.aggregator
        ]
        if len(missing_aggregators) != 0:
            errors.append(
                f"EmbarrassinglyParallelStep '{self.step_name}' has output slots without "
                f"aggregator methods assigned: {missing_aggregators}"
            )
        if errors:
            raise ValueError("\n".join(errors))


class ChoiceStep(Step):
    """A type of :class:`Step` that allows for choosing between multiple paths.

    A ``ChoiceStep`` allows a user to select a single path from a set of possible
    paths.

    See :class:`Step` for inherited attributes.

    Parameters
    ----------
    step_name
        The name of the ``ChoiceStep``.
    input_slots
        All required :class:`InputSlots<easylink.graph_components.InputSlot>`.
    output_slots
        All required :class:`OutputSlots<easylink.graph_components.OutputSlot>`.
    choices
        A dictionary of choices, where the keys are the names/types of choices and
        the values are dictionaries containing that type's nodes, edges, and
        :class:`SlotMappings<easylink.graph_components.SlotMapping>`.

    Notes
    -----
    ``ChoiceSteps`` are by definition non-leaf but do *not* require the typical
    :attr:`Step.config_key` in the pipeline specification file. Instead, the pipeline
    configuration must contain a 'type' key that specifies which option to choose.

    """

    def __init__(
        self,
        step_name: str,
        input_slots: Iterable[InputSlot],
        output_slots: Iterable[OutputSlot],
        choices: dict[
            str, dict[str, list[Step | EdgeParams | InputSlotMapping | OutputSlotMapping]]
        ],
    ) -> None:
        super().__init__(
            step_name,
            input_slots=input_slots,
            output_slots=output_slots,
        )
        self.choices = choices
        """A dictionary of choices, where the keys are the names/types of choices and 
        the values are dictionaries containing that type's nodes, edges, and
        :class:`SlotMappings<easylink.graph_components.SlotMapping>`."""

    def validate_step(
        self,
        step_config: LayeredConfigTree,
        combined_implementations: LayeredConfigTree,
        input_data_config: LayeredConfigTree,
    ) -> dict[str, list[str]]:
        """Validates the ``ChoiceStep``.

        Parameters
        ----------
        step_config
            The configuration of this ``ChoiceStep``.
        combined_implementations
            The configuration for any implementations to be combined.
        input_data_config
            The input data configuration for the entire pipeline.

        Returns
        -------
            A dictionary of errors, where the keys are the ``ChoiceStep`` name and the
            values are lists of error messages associated with the given ``Step``.

        Notes
        -----
        A ``ChoiceStep`` by definition must be set with a :class:`NonLeafConfigurationState`.

        If the ``Step`` does not validate (i.e. errors are found and the returned
        dictionary is non-empty), the tool will exit and the pipeline will not run.

        We attempt to batch error messages as much as possible, but there may be
        times where the configuration is so ill-formed that we are unable to handle
        all issues in one pass. In these cases, new errors may be found after the
        initial ones are handled.

        We update the :class:`easylink.graph_components.StepGraph` and ``SlotMappings``
        in :meth:`validate_step` (as opposed to in :meth:`set_configuration_state`
        as is done in :class:`TemplatedStep`) because :meth:`validate_step` is called
        prior to :meth:`set_configuration_state`, but the validations itself actually
        requires the updated ``StepGraph`` and ``SlotMappings``.

        We do not attempt to validate the subgraph here if the 'type' key is unable
        to be validated.
        """

        chosen_type = step_config.get("type")
        # Handle problems with the 'type' key
        if not chosen_type:
            return {f"step {self.name}": ["The step requires a 'type' key."]}
        if chosen_type not in self.choices:
            return {
                f"step {self.name}": [
                    f"'{step_config['type']}' is not a supported 'type'. Valid choices are: {list(self.choices)}."
                ]
            }
        # Handle type-subgraph inconsistencies
        subgraph = self.choices[chosen_type]
        chosen_step_config = LayeredConfigTree(
            {key: value for key, value in step_config.items() if key != "type"}
        )
        allowable_steps = [node.name for node in subgraph["nodes"]]
        if set(allowable_steps) != set(chosen_step_config):
            return {
                f"step {self.name}": [
                    f"Invalid configuration for '{chosen_type}' type. Valid steps are {allowable_steps}."
                ]
            }

        # HACK: Update the step graph and mappings here because we need them for validation
        self.step_graph = self._update_step_graph(subgraph)
        self.slot_mappings = self._update_slot_mappings(subgraph)
        # NOTE: A ChoiceStep is by definition non-leaf step
        return self._validate_nonleaf(
            chosen_step_config, combined_implementations, input_data_config
        )

    def set_configuration_state(
        self,
        parent_config: LayeredConfigTree,
        combined_implementations: LayeredConfigTree,
        input_data_config: LayeredConfigTree,
    ):
        """Sets the configuration state for a ``ChoiceStep``.

        Parameters
        ----------
        parent_config
            The configuration of the parent ``Step``.
        combined_implementations
            The configuration for any implementations to be combined.
        input_data_config
            The input data configuration for the entire pipeline.

        Notes
        -----
        We update the :class:`easylink.graph_components.StepGraph` and ``SlotMappings``
        in :meth:`validate_step` (as opposed to in :meth:`set_configuration_state`
        as is done in :class:`TemplatedStep`) because :meth:`validate_step` is called
        prior to :meth:`set_configuration_state`, but the validations itself actually
        requires the updated ``StepGraph`` and ``SlotMappings``.
        """

        chosen_parent_config = LayeredConfigTree(
            {key: value for key, value in parent_config[self.name].items() if key != "type"}
        )
        # ChoiceSteps by definition cannot be in a LeafConfigurationState.
        self._configuration_state = NonLeafConfigurationState(
            self, chosen_parent_config, combined_implementations, input_data_config
        )

    @staticmethod
    def _update_step_graph(subgraph: dict[str, Any]) -> StepGraph:
        """Updates the :class:`~easylink.graph_components.StepGraph` with the choice.

        Parameters
        ----------
        subgraph
            Subgraph parameters (nodes, edges, and slot mappings) for the chosen type.

        Returns
        -------
            The updated ``StepGraph`` for the chosen type.
        """
        nodes = subgraph["nodes"]
        edges = subgraph["edges"]

        graph = StepGraph()
        for node in nodes:
            graph.add_node_from_step(node)
        for edge in edges:
            graph.add_edge_from_params(edge)
        return graph

    @staticmethod
    def _update_slot_mappings(subgraph: dict[str, Any]) -> dict[str, list[SlotMapping]]:
        """Updates the :class:`SlotMappings<easylink.graph_components.SlotMapping>` to the choice type.

        Parameters
        ----------
        sub_graph
            Subgraph parameters (nodes, edges, and slot mappings) for the chosen type.

        Returns
        -------
            Updated ``SlotMappings`` that match the choice type.
        """
        input_mappings = subgraph["input_slot_mappings"]
        output_mappings = subgraph["output_slot_mappings"]
        return {"input": input_mappings, "output": output_mappings}


class ConfigurationState(ABC):
    """A given :class:`Step's<Step>` configuration state.

    A ``ConfigurationState`` defines the exact pipeline configuration state for a
    given ``Step``, including the strategy required to get the :class:`~easylink.graph_components.ImplementationGraph`
    from it. There are two possible types of configuration states, "leaf" and "non-leaf",
    and each has its own concrete class, :class:`LeafConfigurationState` and
    ``NonLeafConfigurationState``, respectively.

    Parameters
    ----------
    step
        The ``Step`` this ``ConfigurationState`` is tied to.
    pipeline_config
        The relevant configuration for the ``Step`` we are setting the state for.
    combined_implementations
        The configuration for any implementations to be combined.
    input_data_config
        The input data configuration for the entire pipeline.

    """

    def __init__(
        self,
        step: Step,
        pipeline_config: LayeredConfigTree,
        combined_implementations: LayeredConfigTree,
        input_data_config: LayeredConfigTree,
    ):
        self._step = step
        """The ``Step`` this ``ConfigurationState`` is tied to."""
        self.pipeline_config = pipeline_config
        """The relevant configuration for the ``Step`` we are setting the state for."""
        self.combined_implementations = combined_implementations
        """The relevant configuration if the ``Step's`` ``Implementation``
        has been requested to be combined with that of a different ``Step``."""
        self.input_data_config = input_data_config
        """The input data configuration for the entire pipeline."""

    @abstractmethod
    def get_implementation_graph(self) -> ImplementationGraph:
        """Resolves the graph composed of ``Steps`` into one composed of ``Implementations``."""
        pass

    @abstractmethod
    def get_implementation_edges(self, edge: EdgeParams) -> list[EdgeParams]:
        """Gets the edge information for the ``Implementation`` related to this ``Step``.

        Parameters
        ----------
        edge
            The ``Step's`` edge information to be propagated to the ``ImplementationGraph``.

        Returns
        -------
            The ``Implementation's`` edge information.
        """
        pass


class LeafConfigurationState(ConfigurationState):
    """The :class:`ConfigurationState` for a leaf :class:`Step`.

    A ``LeafConfigurationState`` is a concrete class that corresponds to a leaf
    ``Step``, i.e. one that is implemented by a single :class:`~easylink.implementation.Implementation`.

    See :class:`ConfigurationState` for inherited attributes.

    """

    @property
    def is_combined(self) -> bool:
        """Whether or not this ``Step`` is combined with another ``Step``."""
        return True if COMBINED_IMPLEMENTATION_KEY in self.pipeline_config else False

    @property
    def implementation_config(self) -> LayeredConfigTree:
        """The ``Step's`` specific ``Implementation`` configuration."""
        return (
            self.combined_implementations[self.pipeline_config[COMBINED_IMPLEMENTATION_KEY]]
            if self.is_combined
            else self.pipeline_config["implementation"]
        )

    def get_implementation_graph(self) -> ImplementationGraph:
        """Gets this ``Step's`` :class:`~easylink.graph_components.ImplementationGraph`.

        A ``Step`` in a leaf configuration state by definition has no sub-``Steps``
        to unravel; we are able to directly instantiate an :class:`~easylink.implementation.Implementation`
        and generate an ``ImplementationGraph`` from it.

        Returns
        -------
            The ``ImplementationGraph`` related to this ``Step``.
        """

        implementation_graph = ImplementationGraph()
        implementation_node_name = self._step.implementation_node_name
        if self.is_combined:
            if isinstance(self._step, EmbarrassinglyParallelStep):
                raise NotImplementedError(
                    "Combining implementations with embarrassingly parallel steps "
                    "is not yet supported."
                )
            implementation = PartialImplementation(
                combined_name=self.pipeline_config[COMBINED_IMPLEMENTATION_KEY],
                schema_step=self._step.step_name,
                input_slots=self._step.input_slots.values(),
                output_slots=self._step.output_slots.values(),
            )
        else:
            implementation = Implementation(
                schema_steps=[self._step.step_name],
                implementation_config=self.implementation_config,
                input_slots=self._step.input_slots.values(),
                output_slots=self._step.output_slots.values(),
                is_embarrassingly_parallel=isinstance(self._step, EmbarrassinglyParallelStep),
            )
        implementation_graph.add_node_from_implementation(
            implementation_node_name,
            implementation=implementation,
        )
        return implementation_graph

    def get_implementation_edges(self, edge: EdgeParams) -> list[EdgeParams]:
        """Gets the edge information for the ``Implementation`` related to this ``Step``.

        Parameters
        ----------
        edge
            The ``Step's`` edge information to be propagated to the ``ImplementationGraph``.

        Raises
        ------
        ValueError
            If the ``Step`` is not in the edge or if no edges related to this ``Step`` are found.

        Returns
        -------
            The ``Implementation's`` edge information.
        """
        implementation_edges = []
        if edge.source_node == self._step.name:
            mappings = [
                mapping
                for mapping in self._step.get_implementation_slot_mappings()["output"]
                if mapping.parent_slot == edge.output_slot
            ]
            for mapping in mappings:
                imp_edge = mapping.remap_edge(edge)
                implementation_edges.append(imp_edge)
        elif edge.target_node == self._step.name:
            mappings = [
                mapping
                for mapping in self._step.get_implementation_slot_mappings()["input"]
                if mapping.parent_slot == edge.input_slot
            ]
            for mapping in mappings:
                # FIXME [MIC-5771]: Fix ParallelSteps
                if (
                    "input_data_file" in self.pipeline_config
                    and edge.source_node == "pipeline_graph_input_data"
                ):
                    edge.output_slot = self.pipeline_config["input_data_file"]
                imp_edge = mapping.remap_edge(edge)
                implementation_edges.append(imp_edge)
        else:
            raise ValueError(f"Step {self._step.name} not in edge {edge}")
        if not implementation_edges:
            raise ValueError(f"No edges found for Step {self._step.name} in edge {edge}")
        return implementation_edges


class NonLeafConfigurationState(ConfigurationState):
    """The :class:`ConfigurationState` for a non-leaf :class:`Step`.

    A ``NonLeafConfigurationState`` is a concrete class that corresponds to a non-leaf
    ``Step``, i.e. one that has a non-trivial :class:`~easylink.graph_components.StepGraph`.

    See :class:`ConfigurationState` for inherited attributes.

    Parameters
    ----------
    step
        The ``Step`` this ``ConfigurationState`` is tied to.
    pipeline_config
        The relevant configuration for the ``Step`` we are setting the state for.
    combined_implementations
        The configuration for any implementations to be combined.
    input_data_config
        The input data configuration for the entire pipeline.

    Raises
    ------
    ValueError
        If the ``Step`` does not have a ``StepGraph``.

    Notes
    -----
    The first instance of a ``NonLeafConfigurationState`` is created when calling
    :meth:`~easylink.pipeline_schema.PipelineSchema.configure_pipeline` on the
    :class:`~easylink.pipeline_schema.PipelineSchema` that is chosen for a given
    EasyLink run; the ``step`` passed in is the entire ``PipelineSchema`` and the
    ``pipeline_config`` is that of the entire requested pipeline (which is by definition
    a non-leaf ``Step``).

    Upon instantiation of a ``NonLeafConfigurationState``, the
    :meth:`_configure_subgraph_steps` method is called which iterates through the
    ``Step's`` children and sets their configuration state. If any of these child
    ``Steps`` are also non-leaf, the process continues recursively until all
    nodes are leaf ``Steps`` with a corresponding :class:`LeafConfigurationState`.

    """

    def __init__(
        self,
        step: Step,
        pipeline_config: LayeredConfigTree,
        combined_implementations: LayeredConfigTree,
        input_data_config: LayeredConfigTree,
    ):
        super().__init__(step, pipeline_config, combined_implementations, input_data_config)
        if not step.step_graph:
            raise ValueError(
                "NonLeafConfigurationState requires a subgraph upon which to operate, "
                f"but Step {step.name} has no step graph."
            )
        self._configure_subgraph_steps()

    def get_implementation_graph(self) -> ImplementationGraph:
        """Gets this ``Step's`` :class:`~easylink.graph_components.ImplementationGraph`.

        A ``Step`` in a non-leaf configuration state by definition has a ``StepGraph``
        containing sub-``Steps`` that  need to be unrolled. This method recursively
        traverses that ``StepGraph`` and its childrens' ``StepGraphs`` until all
        sub-``Steps`` are in a :class:`LeafConfigurationState`, i.e. all ``Steps``
        are implemented by a single ``Implementation`` and we have our desired
        ``ImplementationGraph``.

        Returns
        -------
            The ``ImplementationGraph`` of this ``Step``.

        Notes
        -----
        This method is first called on the entire :class:`~easylink.pipeline_schema.PipelineSchema`
        when constructing the :class:`~easylink.pipeline_graph.PipelineGraph`
        to run.

        """
        implementation_graph = ImplementationGraph()
        self.add_nodes(implementation_graph)
        self.add_edges(implementation_graph)
        return implementation_graph

    def add_nodes(self, implementation_graph: ImplementationGraph) -> None:
        """Adds nodes for each ``Step`` to the ``ImplementationGraph``."""
        for node in self._step.step_graph.nodes:
            step = self._step.step_graph.nodes[node]["step"]
            implementation_graph.update(step.get_implementation_graph())

    def add_edges(self, implementation_graph: ImplementationGraph) -> None:
        """Adds the edges to the ``ImplementationGraph``."""
        for source, target, edge_attrs in self._step.step_graph.edges(data=True):
            all_edges = []
            edge = EdgeParams.from_graph_edge(source, target, edge_attrs)
            parent_source_step = self._step.step_graph.nodes[source]["step"]
            parent_target_step = self._step.step_graph.nodes[target]["step"]

            source_edges = parent_source_step.get_implementation_edges(edge)
            for source_edge in source_edges:
                for target_edge in parent_target_step.get_implementation_edges(source_edge):
                    all_edges.append(target_edge)

            for edge in all_edges:
                implementation_graph.add_edge_from_params(edge)

    def get_implementation_edges(self, edge: EdgeParams) -> list[EdgeParams]:
        """Gets the edge information for the ``Implementation`` related to this ``Step``.

        Parameters
        ----------
        edge
            The ``Step's`` edge information to be propagated to the ``ImplementationGraph``.

        Raises
        ------
        ValueError
            If the ``Step`` is not in the edge or if no edges related to this ``Step`` are found.

        Returns
        -------
            The ``Implementation's`` edge information.
        """
        implementation_edges = []
        if edge.source_node == self._step.name:
            mappings = [
                mapping
                for mapping in self._step.slot_mappings["output"]
                if mapping.parent_slot == edge.output_slot
            ]
            for mapping in mappings:
                new_edge = mapping.remap_edge(edge)
                new_step = self._step.step_graph.nodes[mapping.child_node]["step"]
                imp_edges = new_step.get_implementation_edges(new_edge)
                implementation_edges.extend(imp_edges)
        elif edge.target_node == self._step.name:
            mappings = [
                mapping
                for mapping in self._step.slot_mappings["input"]
                if mapping.parent_slot == edge.input_slot
            ]
            for mapping in mappings:
                new_edge = mapping.remap_edge(edge)
                new_step = self._step.step_graph.nodes[mapping.child_node]["step"]
                imp_edges = new_step.get_implementation_edges(new_edge)
                implementation_edges.extend(imp_edges)
        else:
            raise ValueError(f" {self._step.name} not in edge {edge}")
        if not implementation_edges:
            raise ValueError(f"No edges found for {self._step.name} in edge {edge}")
        return implementation_edges

    def _configure_subgraph_steps(self) -> None:
        """Sets the configuration state for all ``Steps`` in the ``StepGraph``.

        This method recursively traverses the ``StepGraph`` and sets the configuration
        state for each ``Step`` until reaching all leaf nodes.
        """
        nodes = self._step.step_graph.nodes
        for node in nodes:
            step = nodes[node]["step"]
            step.set_configuration_state(
                self.pipeline_config, self.combined_implementations, self.input_data_config
            )



# mypy: ignore-errors
import subprocess
import sys
import tempfile
from pathlib import Path

import pytest

from tests.conftest import RESULTS_DIR, SPECIFICATIONS_DIR


@pytest.mark.slow
@pytest.mark.parametrize(
    "pipeline_specification, implementations",
    [
        # Basic
        (
            "common/pipeline.yaml",
            [
                "step_1_python_pandas",
                "step_2_python_pandas",
                "step_3_python_pandas",
                "step_4_python_pandas",
            ],
        ),
        # Expanded pipeline with loops, parallel splits, hierarchy
        (
            "e2e/pipeline_expanded.yaml",
            [
                "step_1_parallel_split_1_step_1_python_pandas",
                "step_1_parallel_split_2_step_1_python_pandas",
                "step_1_parallel_split_3_step_1_python_pandas",
                "step_2_python_pandas",
                "step_3_loop_1_step_3_python_pandas",
                "step_3_loop_2_step_3_python_pandas",
                "step_4a_python_pandas",
                "step_4b_python_pandas",
            ],
        ),
    ],
)
def test_step_types(pipeline_specification, implementations, capsys):
    """Tests against various permutations of complex step types.

    The goal is to test that EasyLink generates the correct output implementations
    depending on the configuration; i.e. if we have a 'substeps' key in the config,
    we get an implementation for each (or else we get a single implementation).
    """
    # Create a temporary directory to store results. We cannot use pytest's tmp_path fixture
    # because other nodes do not have access to it. Also, do not use a context manager
    # (i.e. tempfile.TemporaryDirectory) because it's too difficult to debug when the test
    # fails b/c the dir gets deleted.
    results_dir = tempfile.mkdtemp(dir=RESULTS_DIR)
    input_data = "common/input_data.yaml"
    computing_environment = "common/environment_local.yaml"
    # give the tmpdir the same permissions as the parent directory so that
    # cluster jobs can write to it
    results_dir = Path(results_dir)
    with capsys.disabled():  # disabled so we can monitor job submissions
        print("\n\n*** RUNNING TEST ***\n" f"[{pipeline_specification}]\n")

        cmd = (
            "easylink run "
            f"-p {SPECIFICATIONS_DIR / pipeline_specification} "
            f"-i {SPECIFICATIONS_DIR / input_data} "
            f"-e {SPECIFICATIONS_DIR / computing_environment} "
            f"-o {str(results_dir)} "
            "--no-timestamp"
        )
        subprocess.run(
            cmd,
            shell=True,
            stdout=sys.stdout,
            stderr=sys.stderr,
            check=True,
        )
        final_output = results_dir / "result.parquet"
        assert final_output.exists()

        # Check that we get directories for particular implementations
        diagnostics_dir = results_dir / "diagnostics"
        for implementation in implementations:
            assert (diagnostics_dir / implementation).exists()
            assert (results_dir / "intermediate" / implementation / "result.parquet").exists()
        print("\n\n*** END OF TEST ***\n" f"[{pipeline_specification}]\n")



"""
==============
Pipeline Graph
==============

This module is responsible for managing the directed acyclic graph (DAG) of the
entire pipeline to be run from an ``easylink run`` call.

"""

import itertools
from collections import Counter, defaultdict
from copy import deepcopy
from pathlib import Path

import networkx as nx

from easylink.configuration import Config
from easylink.graph_components import (
    EdgeParams,
    ImplementationGraph,
    InputSlot,
    InputSlotMapping,
    OutputSlot,
    OutputSlotMapping,
)
from easylink.implementation import Implementation, PartialImplementation
from easylink.utilities import paths
from easylink.utilities.data_utils import load_yaml


class PipelineGraph(ImplementationGraph):
    """A directed acyclic graph (DAG) of the pipeline to run.

    The ``PipelineGraph`` is a DAG of the entire pipeline to run. It has
    :class:`Implementations<easylink.implementation.Implementation>` for nodes
    and the file dependencies between them for edges. Multiple edges between nodes
    are permitted.

    This is the highest level type of :class:`~easylink.graph_components.ImplementationGraph`.

    See :class:`~easylink.graph_components.ImplementationGraph` for inherited attributes.

    Parameters
    ----------
    config
        The :class:`~easylink.configuration.Config` object.
    freeze
        Whether to freeze the graph after construction.

    Notes
    -----
    The ``PipelineGraph`` is a low-level abstraction; it represents the *actual
    implementations* of each :class:`~easylink.step.Step` in the resolved pipeline
    being run. This is in contrast to the :class:`~easylink.pipeline_schema.PipelineSchema`,
    which can be an intricate nested structure due to the various complex and self-similar
    ``Step`` instances (which represent abstract operations  such as "loop this
    step N times"). A ``PipelineGraph`` is the flattened and concrete graph of
    ``Implementations`` to run.
    """

    def __init__(self, config: Config, freeze: bool = True) -> None:
        super().__init__(incoming_graph_data=config.schema.get_implementation_graph())
        self._merge_combined_implementations(config)
        self._update_slot_filepaths(config)
        if freeze:
            self = nx.freeze(self)

    @property
    def spark_is_required(self) -> bool:
        """Whether or not any :class:`~easylink.implementation.Implementation` requires spark."""
        return any([implementation.requires_spark for implementation in self.implementations])

    @property
    def any_embarrassingly_parallel(self) -> bool:
        """Whether or not any :class:`~easylink.implementation.Implementation` is
        to be run in an embarrassingly parallel way."""
        return any(
            [
                self.get_whether_embarrassingly_parallel(node)
                for node in self.implementation_nodes
            ]
        )

    def get_whether_embarrassingly_parallel(self, node: str) -> bool:
        """Determines whether a node is to be run in an embarrassingly parallel way.

        Parameters
        ----------
        node
            The node name to determine whether or not it is to be run in an
            embarrassingly parallel way.

        Returns
        -------
            A boolean indicating whether the node is to be run in an embarrassingly
            parallel way.
        """
        return self.nodes[node]["implementation"].is_embarrassingly_parallel

    def get_io_filepaths(self, node: str) -> tuple[list[str], list[str]]:
        """Gets all of a node's input and output filepaths from its edges.

        Parameters
        ----------
        node
            The node name to get input and output filepaths for.

        Returns
        -------
            The input and output filepaths for the node.
        """
        input_files = list(
            itertools.chain.from_iterable(
                [
                    edge_attrs["filepaths"]
                    for _, _, edge_attrs in self.in_edges(node, data=True)
                ]
            )
        )
        output_files = list(
            itertools.chain.from_iterable(
                [
                    edge_attrs["filepaths"]
                    for _, _, edge_attrs in self.out_edges(node, data=True)
                ]
            )
        )
        return input_files, output_files

    def get_io_slot_attributes(
        self, node: str
    ) -> tuple[dict[str, dict[str, str | list[str]]], dict[str, dict[str, str | list[str]]]]:
        """Gets all of a node's i/o slot attributes from edges.

        Parameters
        ----------
        node
            The node name to get slot attributes for.

        Returns
        -------
            A tuple of mappings of node name to slot attributes.
        """
        input_slots = [
            edge_attrs["input_slot"] for _, _, edge_attrs in self.in_edges(node, data=True)
        ]
        input_filepaths_by_slot = [
            list(edge_attrs["filepaths"])
            for _, _, edge_attrs in self.in_edges(node, data=True)
        ]
        input_slot_attrs = self._deduplicate_input_slots(input_slots, input_filepaths_by_slot)

        output_slots = [
            edge_attrs["output_slot"] for _, _, edge_attrs in self.out_edges(node, data=True)
        ]
        output_filepaths_by_slot = [
            list(edge_attrs["filepaths"])
            for _, _, edge_attrs in self.out_edges(node, data=True)
        ]
        output_slot_attrs = self._deduplicate_output_slots(
            output_slots, output_filepaths_by_slot
        )
        return input_slot_attrs, output_slot_attrs

    ##################
    # Helper Methods #
    ##################

    def _merge_combined_implementations(self, config: Config) -> None:
        """Merge nodes with the same combined implementation into a single node.

        It is sometimes useful to use a single :class:`~easylink.implementation.Implementation`
        that implements multiple steps of a pipeline at once. If the user has specified
        such a scenario (via the pipeline configuration), this method will merge
        all nodes that use the same combined implementation into a single node.

        Parameters
        ----------
        config
            The :class:`~easylink.configuration.Config` object.

        Raises
        ------
        ValueError
            If the ``PipelineGraph`` contains a cycle after combining implementations.

        Notes
        -----
        There are no special class abstractions to represent combined implementations.
        Rather, the merging of nodes is explicitly called directly on the original
        ``PipelineGraph``.
        """
        implementation_metadata = load_yaml(paths.IMPLEMENTATION_METADATA)
        for (
            combined_implementation,
            combined_implementation_config,
        ) in config.pipeline.combined_implementations.items():

            # Find all nodes with the same combined implementation
            nodes_to_merge = [
                node
                for node, data in self.nodes(data=True)
                if isinstance(data["implementation"], PartialImplementation)
                and data["implementation"].combined_name == combined_implementation
            ]
            if not nodes_to_merge:
                continue

            partial_implementations_to_merge = [
                self.nodes[node]["implementation"] for node in nodes_to_merge
            ]

            implemented_steps = [
                implementation.schema_step
                for implementation in partial_implementations_to_merge
            ]

            metadata_steps = implementation_metadata[combined_implementation_config["name"]][
                "steps"
            ]
            self._validate_combined_implementation_topology(nodes_to_merge, metadata_steps)

            (
                combined_input_slots,
                combined_output_slots,
                combined_edges,
            ) = self._get_combined_slots_and_edges(combined_implementation, nodes_to_merge)

            # Create new implementation
            new_implementation = Implementation(
                implemented_steps,
                combined_implementation_config,
                combined_input_slots,
                combined_output_slots,
            )
            self.add_node(combined_implementation, implementation=new_implementation)

            # Redirect edges
            for edge in combined_edges:
                self.add_edge_from_params(edge)

            # Remove original nodes
            self.remove_nodes_from(nodes_to_merge)
            try:
                cycle = nx.find_cycle(self)
                raise ValueError(
                    "The pipeline graph contains a cycle after combining "
                    "implementations: {}".format(cycle)
                )
            except nx.NetworkXNoCycle:
                pass

    def _validate_combined_implementation_topology(
        self, nodes: list[str], metadata_steps: list[str]
    ) -> None:
        """Validates the combined implementation topology against intended implementation.

        Parameters
        ----------
        nodes
            Sorted names of the nodes to validate.
        metadata_steps
            Sorted names of the steps that the nodes are intended to implement.

        Raises
        ------
        ValueError
            If there is a mismatch between the nodes and the steps they intend
            to implement or if the nodes are no longer topologically consistent.
        """
        # HACK: We cannot just call self.subgraph(nodes) because networkx will
        # try and instantiate another PipelineGraph which requires a Config object
        # to be passed to the constructor.
        subgraph = ImplementationGraph(self).subgraph(nodes)

        # Relabel nodes by schema step
        mapping = {
            node: data["implementation"].schema_step
            for node, data in subgraph.nodes(data=True)
        }
        if not set(mapping.values()) == set(metadata_steps):
            # NOTE: It's possible that we've combined nodes in such a way that removed
            # an edge from the graph and so nx is unable to reliably sort the subgraph.
            full_pipeline_sorted_nodes = list(nx.topological_sort(self))
            sorted_nodes = [node for node in full_pipeline_sorted_nodes if node in mapping]
            raise ValueError(
                f"Pipeline configuration nodes {[mapping[node] for node in sorted_nodes]} "
                f"do not match metadata steps {metadata_steps}."
            )
        subgraph = nx.relabel_nodes(subgraph, mapping)
        # Check for topological inconsistency, i.e. if there is a path from a later
        # node to an earlier node.
        for predecessor in range(len(metadata_steps)):
            for successor in range(predecessor + 1, len(metadata_steps)):
                if nx.has_path(
                    subgraph, metadata_steps[successor], metadata_steps[predecessor]
                ):
                    raise ValueError(
                        f"Pipeline configuration nodes {list(nx.topological_sort(subgraph))} "
                        "are not topologically consistent with the intended implementations "
                        f"for {list(metadata_steps)}:\n"
                        f"There is a path from successor {metadata_steps[successor]} "
                        f"to predecessor {metadata_steps[predecessor]}."
                    )

    def _get_combined_slots_and_edges(
        self, combined_implementation: str, nodes_to_merge: list[str]
    ) -> tuple[set[InputSlot], set[OutputSlot], set[EdgeParams]]:
        """Gets all edge information required for the combined implementation.

        Parameters
        ----------
        combined_implementation
            The name of the combined implementation.
        nodes_to_merge
            The names of the nodes being merged.

        Returns
        -------
            The :class:`InputSlots<easylink.graph_components.InputSlot>`,
            :class:`OutputSlots<easylink.graph_components.OutputSlot>`, and
            :class:`~easylink.graph_components.EdgeParams` needed to construct the
            combined implementation.

        Notes
        -----
        When combining implementations results in a node with multiple slots with
        the same name and/or environment variable, the slots are made unique
        by prepending the :class:`~easylink.step.Step` name to the slot name as well
        as to the environment variable. This is necessary to prevent collisions
        with a combined implementation that takes multiple environment variables that
        have the same name.
        """
        slot_types = ["input_slot", "output_slot"]
        combined_slots_by_type = combined_input_slots, combined_output_slots = set(), set()
        edges_by_slot_and_type = self._get_edges_by_slot(nodes_to_merge)
        transform_mappings = (InputSlotMapping, OutputSlotMapping)

        combined_edges = set()
        # FIXME [MIC-5848]: test coverage is lacking when two output slots have the same name,
        # i.e. combing two steps that have the same name output slots
        for slot_type, combined_slots, edges_by_slot, transform_mapping in zip(
            slot_types, combined_slots_by_type, edges_by_slot_and_type, transform_mappings
        ):
            separate_slots = edges_by_slot.keys()
            duplicate_slots = self._get_duplicate_slots(separate_slots, slot_type)
            for slot_tuple in separate_slots:
                step_name, slot = slot_tuple
                edges = edges_by_slot[slot_tuple]
                if slot_tuple in duplicate_slots:
                    combined_slot = slot.__class__(
                        name=step_name + "_" + slot.name,
                        env_var=step_name.upper() + "_" + slot.env_var,
                        validator=slot.validator,
                    )
                else:
                    combined_slot = deepcopy(slot)
                combined_slots.add(combined_slot)
                slot_mapping = transform_mapping(
                    slot.name, combined_implementation, combined_slot.name
                )
                combined_edges.update([slot_mapping.remap_edge(edge) for edge in edges])
        return combined_input_slots, combined_output_slots, combined_edges

    def _get_edges_by_slot(
        self,
        nodes_to_merge: list[str],
    ) -> tuple[
        dict[tuple[str, InputSlot], EdgeParams], dict[tuple[str, OutputSlot], EdgeParams]
    ]:
        """Gets the edge information for all nodes to be combined.

        Parameters
        ----------
        nodes_to_merge
            A list of ``PipelineGraph`` nodes to be combined.

        Returns
        -------
            Input and output slot-edge mappings.
        """

        in_edges_by_slot = defaultdict(list)
        out_edges_by_slot = defaultdict(list)

        # Find separate input and output slots
        for node in nodes_to_merge:
            for pred, succ, data in self.in_edges(node, data=True):
                if pred not in nodes_to_merge:
                    input_slot = data["input_slot"]
                    partial_implementation = self.nodes[node]["implementation"]
                    in_edges_by_slot[(partial_implementation.schema_step, input_slot)].append(
                        EdgeParams.from_graph_edge(pred, succ, data)
                    )

            for pred, succ, data in self.out_edges(node, data=True):

                if succ not in nodes_to_merge:
                    output_slot = data["output_slot"]
                    partial_implementation = self.nodes[node]["implementation"]
                    out_edges_by_slot[
                        (partial_implementation.schema_step, output_slot)
                    ].append(EdgeParams.from_graph_edge(pred, succ, data))
        return in_edges_by_slot, out_edges_by_slot

    @staticmethod
    def _get_duplicate_slots(
        slot_tuples: set[tuple[str, InputSlot | OutputSlot]], slot_type: str
    ) -> set[tuple[str, InputSlot | OutputSlot]]:
        """Gets duplicate slots.

        Combining nodes can lead to duplicate slots. In order to deduplicate them,
        this helper method returns only those which have duplicate names or
        environment variables.

        Parameters
        ----------
        slot_tuples
            A set of (step_name, slot) tuples to check for duplication.
        slot_type
            "input_slot" or "output_slot"

        Returns
        -------
            A set of (step_name, slot) tuples that have duplicate names or environment
            variables to be deduplicated.
        """
        name_freq = Counter([slot.name for step_name, slot in slot_tuples])
        duplicate_names = [name for name, count in name_freq.items() if count > 1]
        if slot_type == "input_slot":
            env_var_freq = Counter([slot.env_var for step_name, slot in slot_tuples])
            duplicate_env_vars = [
                env_var for env_var, count in env_var_freq.items() if count > 1
            ]
            duplicate_slots = {
                (step_name, slot)
                for (step_name, slot) in slot_tuples
                if slot.name in duplicate_names or slot.env_var in duplicate_env_vars
            }
        else:
            duplicate_slots = {
                (step_name, slot)
                for (step_name, slot) in slot_tuples
                if slot.name in duplicate_names
            }
        return duplicate_slots

    def _update_slot_filepaths(self, config: Config) -> None:
        """Fills graph edges with appropriate filepath information.

        This method updates the :class:`~easylink.step.Step` slot information with
        actual filepaths. This can't happen earlier in the process because we
        don't know node names until now (which are required for the filepaths).

        Parameters
        ----------
        config
            The :class:`~easylink.configuration.Config`.
        """
        # Update input data edges to correct filenames from config
        for src, sink, edge_attrs in self.out_edges("input_data", data=True):
            for edge_idx in self[src][sink]:
                if edge_attrs["output_slot"].name == "all":
                    self[src][sink][edge_idx]["filepaths"] = tuple(
                        str(path) for path in config.input_data.to_dict().values()
                    )
                else:
                    self[src][sink][edge_idx]["filepaths"] = (
                        str(config.input_data[edge_attrs["output_slot"].name]),
                    )

        # Update implementation nodes with yaml metadata
        for node in self.implementation_nodes:
            implementation = self.nodes[node]["implementation"]
            imp_outputs = implementation.outputs
            for src, sink, edge_attrs in self.out_edges(node, data=True):
                for edge_idx in self[node][sink]:
                    self[src][sink][edge_idx]["filepaths"] = (
                        str(
                            Path("intermediate")
                            / node
                            / imp_outputs[edge_attrs["output_slot"].name]
                        ),
                    )

    @staticmethod
    def _deduplicate_input_slots(
        input_slots: list[InputSlot], filepaths_by_slot: list[str]
    ) -> dict[str, dict[str, str | list[str]]]:
        """Deduplicates input slots into a dictionary with filepaths.

        Parameters
        ----------
        input_slots
            The :class:`InputSlots<easylink.graph_components.InputSlot>` condense.
        filepaths_by_slot
            The filepaths associated with each ``InputSlot``.

        Returns
        -------
            A dictionary mapping ``InputSlot`` names to their attributes and filepaths.

        Raises
        ------
        ValueError
            If duplicate slot names are found with different environment variables
            or validators.
        """
        condensed_slot_dict = {}
        for input_slot, filepaths in zip(input_slots, filepaths_by_slot):
            slot_name, env_var, validator, splitter = (
                input_slot.name,
                input_slot.env_var,
                input_slot.validator,
                input_slot.splitter,
            )
            if slot_name in condensed_slot_dict:
                if env_var != condensed_slot_dict[slot_name]["env_var"]:
                    raise ValueError(
                        f"Duplicate input slots named '{slot_name}' have different env vars."
                    )
                condensed_slot_validator = condensed_slot_dict[slot_name]["validator"]
                if validator != condensed_slot_validator:
                    raise ValueError(
                        f"Duplicate input slots named '{slot_name}' have different validators: "
                        f"'{validator.__name__}' and '{condensed_slot_validator.__name__}'."
                    )
                # Add the new filepaths to the existing slot
                condensed_slot_dict[slot_name]["filepaths"].extend(filepaths)
            else:
                condensed_slot_dict[slot_name] = {
                    "env_var": env_var,
                    "validator": validator,
                    "filepaths": filepaths,
                    "splitter": splitter,
                }
        return condensed_slot_dict

    @staticmethod
    def _deduplicate_output_slots(
        output_slots: list[OutputSlot], filepaths_by_slot: list[str]
    ) -> dict[str, dict[str, str | list[str]]]:
        """Deduplicates output slots into a dictionary with filepaths.

        Parameters
        ----------
        output_slots
            The :class:`OutputSlots<easylink.graph_components.OutputSlot>` to deduplicate.
        filepaths_by_slot
            The filepaths associated with each ``OutputSlot``.

        Returns
        -------
            A dictionary mapping ``OutputSlot`` names to their attributes and filepaths.
        """
        condensed_slot_dict = {}
        for output_slot, filepaths in zip(output_slots, filepaths_by_slot):
            slot_name, aggregator = (
                output_slot.name,
                output_slot.aggregator,
            )
            if slot_name in condensed_slot_dict:
                # Add the new filepaths to the existing slot
                condensed_slot_dict[slot_name]["filepaths"].extend(filepaths)
            else:
                condensed_slot_dict[slot_name] = {
                    "filepaths": filepaths,
                    "aggregator": aggregator,
                }
        return condensed_slot_dict



"""
=============
Configuration
=============

This module is responsible for managing an easylink run's configuration as defined
by various user-input specification files.

"""

from collections import defaultdict
from pathlib import Path
from typing import Any

from layered_config_tree import LayeredConfigTree

from easylink.pipeline_schema import PIPELINE_SCHEMAS, PipelineSchema
from easylink.utilities.data_utils import load_yaml
from easylink.utilities.general_utils import exit_with_validation_error

PIPELINE_ERRORS_KEY = "PIPELINE ERRORS"
INPUT_DATA_ERRORS_KEY = "INPUT DATA ERRORS"
ENVIRONMENT_ERRORS_KEY = "ENVIRONMENT ERRORS"

DEFAULT_ENVIRONMENT = {
    "environment": {
        "computing_environment": "local",
        "container_engine": "undefined",
        "implementation_resources": {
            "memory": 1,  # GB
            "cpus": 1,
            "time_limit": 1,  # hours
        },
    }
}
"""The default environment configuration settings."""

SPARK_DEFAULTS = {
    "workers": {
        "num_workers": 2,
        "cpus_per_node": 1,
        "mem_per_node": 1,  # GB
        "time_limit": 1,  # hours
    },
    "keep_alive": False,
}
"""The default spark configuration settings."""

SLURM_SPARK_MEM_BUFFER = 500  # MB


class Config(LayeredConfigTree):
    """A container for configuration information.

    The ``Config`` contains the user-provided specifications for the pipeline,
    input data, and computing environment specifications. It is a nested
    dictionary-like object that supports prioritized layers of configuration settings
    as well as dot-notation access to its attributes.

    The ``Config`` is also reponsible for various validation checks on the provided
    specifications. If any of these are invalid, a validation error is raised with
    as much information as can possibly be provided.

    Parameters
    ----------
    config_params
        A dictionary of all specifications required to run the pipeline. This
        includes the pipeline, input data, and computing environment specifications,
        as well as the results directory.
    potential_schemas
        A list of potential schemas to validate the pipeline configuration against.
        This is primarily used for testing purposes. Defaults to the supported schemas.

    Attributes
    ----------
    environment
        The environment configuration, including computing environment,
        container engine, implementation resources, and slurm- and spark-specific
        requests.
    pipeline
        The pipeline configuration.
    input_data
        The input data filepaths.
    schema
        The :class:`~easylink.pipeline_schema.PipelineSchema` that successfully
        validated the requested pipeline.

    Notes
    -----
    The requested pipeline is checked against a set of supported
    ``PipelineSchemas``. The first schema that successfully validates is assumed
    to be the correct one and is attached to the ``Config`` object and its
    :meth:`~easylink.pipeline_schema.PipelineSchema.configure_pipeline`
    method is called.
    """

    def __init__(
        self,
        config_params: dict[str, Any],
        potential_schemas: PipelineSchema | list[PipelineSchema] = PIPELINE_SCHEMAS,
    ) -> None:
        super().__init__(layers=["initial_data", "default", "user_configured"])
        self.update(DEFAULT_ENVIRONMENT, layer="default")
        self.update(config_params, layer="user_configured")
        self.update({"environment": {"spark": SPARK_DEFAULTS}}, layer="default")
        self.update({"pipeline": {"combined_implementations": {}}}, layer="default")
        if self.environment.computing_environment == "slurm":
            # Set slurm defaults to empty dict instead of None so that we don't get errors
            # in slurm_resources property
            self.update({"environment": {"slurm": {}}}, layer="default")
        if not isinstance(potential_schemas, list):
            potential_schemas = [potential_schemas]
        self.update({"schema": self._get_schema(potential_schemas)}, layer="initial_data")
        self.schema.configure_pipeline(self.pipeline, self.input_data)
        self._validate()
        self.freeze()

    @property
    def computing_environment(self) -> str:
        """The computing environment to run on ('local' or 'slurm')."""
        return self.environment.computing_environment

    @property
    def slurm(self) -> dict[str, Any]:
        """A dictionary of slurm-specific configuration settings."""
        if not self.environment.computing_environment == "slurm":
            return {}
        else:
            return self.environment.slurm.to_dict()

    @property
    def spark(self) -> dict[str, Any]:
        """A dictionary of spark-specific configuration settings."""
        return self.environment.spark.to_dict()

    @property
    def slurm_resources(self) -> dict[str, str]:
        """A flat dictionary of slurm resource requests."""
        if not self.computing_environment == "slurm":
            return {}
        raw_slurm_resources = {
            **self.slurm,
            **self.environment.implementation_resources.to_dict(),
        }
        return {
            "slurm_account": f"'{raw_slurm_resources.get('account')}'",
            "slurm_partition": f"'{raw_slurm_resources.get('partition')}'",
            "mem_mb": int(raw_slurm_resources.get("memory", 0) * 1024),
            "runtime": int(raw_slurm_resources.get("time_limit") * 60),
            "cpus_per_task": raw_slurm_resources.get("cpus"),
        }

    @property
    def spark_resources(self) -> dict[str, Any]:
        """A flat dictionary of spark resource requests."""
        spark_workers_raw = self.spark["workers"]
        spark_workers = {
            "num_workers": spark_workers_raw.get("num_workers"),
            "mem_mb": int(spark_workers_raw.get("mem_per_node", 0) * 1024),
            "slurm_mem_mb": int(
                spark_workers_raw.get("mem_per_node", 0) * 1024 + SLURM_SPARK_MEM_BUFFER
            ),
            "runtime": int(spark_workers_raw.get("time_limit") * 60),
            "cpus_per_task": spark_workers_raw.get("cpus_per_node"),
        }
        return {
            **self.slurm,
            **spark_workers,
            **{k: v for k, v in self.spark.items() if k != "workers"},
        }

    #################
    # Setup Methods #
    #################

    def _get_schema(self, potential_schemas: list[PipelineSchema]) -> PipelineSchema:
        """Returns the first :class:`~easylink.pipeline_schema.PipelineSchema` that validates the requested pipeline.

        Parameters
        ----------
        potential_schemas
            ``PipelineSchemas`` to validate the pipeline configuration against.

        Returns
        -------
            The first ``PipelineSchema`` that validates the requested pipeline configuration.

        Raises
        ------
        SystemExit
            If the pipeline configuration is not valid for any of the ``potential_schemas``,
            the program exits with a non-zero code and all validation errors found
            are logged.

        Notes
        -----
        This acts as the pipeline configuration file's validation method since
        we can only find a matching ``PipelineSchema`` if that file is valid.

        This method returns the *first* ``PipelineSchema`` that validates and does
        not attempt to check additional ones.
        """
        errors = defaultdict(dict)
        # Try each schema until one is validated
        for schema in potential_schemas:
            logs = schema.validate_step(self.pipeline, self.input_data)
            if logs:
                errors[PIPELINE_ERRORS_KEY][schema.name] = logs
                pass  # try the next schema
            else:  # schema was validated
                return schema
        # No schemas were validated
        exit_with_validation_error(dict(errors))

    def _validate(self) -> None:
        """Validates the ``Config``.

        Raises
        ------
        SystemExit
            If any errors are found, they are batch-logged into a dictionary and
            the program exits with a non-zero code.

        Notes
        -----
        Pipeline validations are handled in :meth:`~easylink.configuration.Config._get_schema`.
        """
        # TODO [MIC-4880]: refactor into validation object
        errors = {
            # NOTE: pipeline configuration validation happens in '_get_schema()'
            **self._validate_input_data(),
            **self._validate_environment(),
        }
        if errors:
            exit_with_validation_error(errors)

    def _validate_input_data(self) -> dict[Any, Any]:
        """Validates the input data configuration.

        Returns
        -------
            A dictionary of input data configuration validation errors.
        """
        errors = defaultdict(dict)
        input_data_dict = self.input_data.to_dict()
        if not input_data_dict:
            errors[INPUT_DATA_ERRORS_KEY] = ["No input data is configured."]
        else:
            input_data_errors = self.schema.validate_inputs(input_data_dict)
            if input_data_errors:
                errors[INPUT_DATA_ERRORS_KEY] = input_data_errors
        return errors

    def _validate_environment(self) -> dict[Any, Any]:
        """Validates the environment configuration.

        Returns
        -------
            A dictionary of environment configuration validation errors.
        """
        errors = defaultdict(dict)
        if not self.environment.container_engine in ["docker", "singularity", "undefined"]:
            errors[ENVIRONMENT_ERRORS_KEY]["container_engine"] = [
                f"The value '{self.environment.container_engine}' is not supported."
            ]

        if self.environment.computing_environment == "slurm" and not self.environment.slurm:
            errors[ENVIRONMENT_ERRORS_KEY]["slurm"] = [
                "The environment configuration file must include a 'slurm' key "
                "defining slurm resources if the computing_environment is 'slurm'."
            ]

        return errors


def load_params_from_specification(
    pipeline_specification: str,
    input_data: str,
    computing_environment: str | None,
    results_dir: str,
) -> dict[str, Any]:
    """Gathers together all specification data.

    This gathers the pipeline, input data, and computing environment specifications
    as well as the results directory into a single dictionary for insertion into
    the :class:`Config` object.

    Parameters
    ----------
    pipeline_specification
        The path to the pipeline specification file.
    input_data
        The path to the input data specification file.
    computing_environment
        The path to the computing environment specification file.
    results_dir
        The path to the results directory.

    Returns
    -------
        A dictionary of all provided specification data.
    """
    return {
        "pipeline": load_yaml(pipeline_specification),
        "input_data": _load_input_data_paths(input_data),
        "environment": _load_computing_environment(computing_environment),
        "results_dir": Path(results_dir),
    }


def _load_input_data_paths(
    input_data_specification_path: str | Path,
) -> dict[str, list[Path]]:
    """Creates a dictionary of input data paths from the input data specification file."""
    input_data_paths = load_yaml(input_data_specification_path)
    if not isinstance(input_data_paths, dict):
        raise TypeError(
            "Input data should be submitted like 'key': path/to/file. "
            f"Input was: '{input_data_paths}'"
        )
    filepath_dict = {
        filename: Path(filepath).resolve() for filename, filepath in input_data_paths.items()
    }
    return filepath_dict


def _load_computing_environment(
    computing_environment_specification_path: str | None,
) -> dict[Any, Any]:
    """Loads the computing environment specification file and returns the contents as a dict."""
    if not computing_environment_specification_path:
        return {}  # handles empty environment.yaml
    elif not Path(computing_environment_specification_path).is_file():
        raise FileNotFoundError(
            "Computing environment is expected to be a path to an existing"
            f" specification file. Input was: '{computing_environment_specification_path}'"
        )
    else:
        return load_yaml(computing_environment_specification_path)



"""
============================
Development Pipeline Schemas
============================
"""

from easylink.graph_components import (
    EdgeParams,
    InputSlot,
    InputSlotMapping,
    OutputSlot,
    OutputSlotMapping,
)
from easylink.step import (
    ChoiceStep,
    EmbarrassinglyParallelStep,
    HierarchicalStep,
    InputStep,
    LoopStep,
    OutputStep,
    ParallelStep,
    Step,
)
from easylink.utilities.aggregator_utils import concatenate_datasets
from easylink.utilities.splitter_utils import split_data_by_size
from easylink.utilities.validation_utils import validate_input_file_dummy

NODES = [
    InputStep(),
    ParallelStep(
        template_step=Step(
            step_name="step_1",
            input_slots=[
                InputSlot(
                    name="step_1_main_input",
                    env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validator=validate_input_file_dummy,
                ),
            ],
            output_slots=[OutputSlot("step_1_main_output")],
        ),
    ),
    Step(
        step_name="step_2",
        input_slots=[
            InputSlot(
                name="step_2_main_input",
                env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                validator=validate_input_file_dummy,
            )
        ],
        output_slots=[OutputSlot("step_2_main_output")],
    ),
    LoopStep(
        template_step=EmbarrassinglyParallelStep(
            step_name="step_3",
            input_slots=[
                InputSlot(
                    name="step_3_main_input",
                    env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validator=validate_input_file_dummy,
                    splitter=split_data_by_size,
                ),
            ],
            output_slots=[
                OutputSlot(
                    name="step_3_main_output",
                    aggregator=concatenate_datasets,
                ),
            ],
        ),
        self_edges=[
            EdgeParams(
                source_node="step_3",
                target_node="step_3",
                output_slot="step_3_main_output",
                input_slot="step_3_main_input",
            )
        ],
    ),
    ChoiceStep(
        step_name="choice_section",
        input_slots=[
            InputSlot(
                name="choice_section_main_input",
                env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                validator=validate_input_file_dummy,
            ),
            InputSlot(
                name="choice_section_secondary_input",
                env_var="DUMMY_CONTAINER_SECONDARY_INPUT_FILE_PATHS",
                validator=validate_input_file_dummy,
            ),
        ],
        output_slots=[OutputSlot("choice_section_main_output")],
        choices={
            "simple": {
                "nodes": [
                    HierarchicalStep(
                        step_name="step_4",
                        input_slots=[
                            InputSlot(
                                name="step_4_main_input",
                                env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                                validator=validate_input_file_dummy,
                            ),
                            InputSlot(
                                name="step_4_secondary_input",
                                env_var="DUMMY_CONTAINER_SECONDARY_INPUT_FILE_PATHS",
                                validator=validate_input_file_dummy,
                            ),
                        ],
                        output_slots=[OutputSlot("step_4_main_output")],
                        nodes=[
                            Step(
                                step_name="step_4a",
                                input_slots=[
                                    InputSlot(
                                        name="step_4a_main_input",
                                        env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                                        validator=validate_input_file_dummy,
                                    ),
                                    InputSlot(
                                        name="step_4a_secondary_input",
                                        env_var="DUMMY_CONTAINER_SECONDARY_INPUT_FILE_PATHS",
                                        validator=validate_input_file_dummy,
                                    ),
                                ],
                                output_slots=[OutputSlot("step_4a_main_output")],
                            ),
                            Step(
                                step_name="step_4b",
                                input_slots=[
                                    InputSlot(
                                        name="step_4b_main_input",
                                        env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                                        validator=validate_input_file_dummy,
                                    ),
                                    InputSlot(
                                        name="step_4b_secondary_input",
                                        env_var="DUMMY_CONTAINER_SECONDARY_INPUT_FILE_PATHS",
                                        validator=validate_input_file_dummy,
                                    ),
                                ],
                                output_slots=[OutputSlot("step_4b_main_output")],
                            ),
                        ],
                        edges=[
                            EdgeParams(
                                source_node="step_4a",
                                target_node="step_4b",
                                output_slot="step_4a_main_output",
                                input_slot="step_4b_main_input",
                            ),
                        ],
                        input_slot_mappings=[
                            InputSlotMapping(
                                parent_slot="step_4_main_input",
                                child_node="step_4a",
                                child_slot="step_4a_main_input",
                            ),
                            InputSlotMapping(
                                parent_slot="step_4_secondary_input",
                                child_node="step_4a",
                                child_slot="step_4a_secondary_input",
                            ),
                            InputSlotMapping(
                                parent_slot="step_4_secondary_input",
                                child_node="step_4b",
                                child_slot="step_4b_secondary_input",
                            ),
                        ],
                        output_slot_mappings=[
                            OutputSlotMapping(
                                parent_slot="step_4_main_output",
                                child_node="step_4b",
                                child_slot="step_4b_main_output",
                            ),
                        ],
                    ),
                ],
                "edges": [],
                "input_slot_mappings": [
                    InputSlotMapping(
                        parent_slot="choice_section_main_input",
                        child_node="step_4",
                        child_slot="step_4_main_input",
                    ),
                    InputSlotMapping(
                        parent_slot="choice_section_secondary_input",
                        child_node="step_4",
                        child_slot="step_4_secondary_input",
                    ),
                ],
                "output_slot_mappings": [
                    OutputSlotMapping(
                        parent_slot="choice_section_main_output",
                        child_node="step_4",
                        child_slot="step_4_main_output",
                    ),
                ],
            },
            "complex": {
                "nodes": [
                    Step(
                        step_name="step_5",
                        input_slots=[
                            InputSlot(
                                name="step_5_main_input",
                                env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                                validator=validate_input_file_dummy,
                            ),
                        ],
                        output_slots=[OutputSlot("step_5_main_output")],
                    ),
                    Step(
                        step_name="step_6",
                        input_slots=[
                            InputSlot(
                                name="step_6_main_input",
                                env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                                validator=validate_input_file_dummy,
                            ),
                        ],
                        output_slots=[OutputSlot("step_6_main_output")],
                    ),
                ],
                "edges": [
                    EdgeParams(
                        source_node="step_5",
                        target_node="step_6",
                        output_slot="step_5_main_output",
                        input_slot="step_6_main_input",
                    ),
                ],
                "input_slot_mappings": [
                    InputSlotMapping(
                        parent_slot="choice_section_main_input",
                        child_node="step_5",
                        child_slot="step_5_main_input",
                    ),
                ],
                "output_slot_mappings": [
                    OutputSlotMapping(
                        parent_slot="choice_section_main_output",
                        child_node="step_6",
                        child_slot="step_6_main_output",
                    ),
                ],
            },
        },
    ),
    OutputStep(
        input_slots=[
            InputSlot(name="result", env_var=None, validator=validate_input_file_dummy)
        ],
    ),
]
EDGES = [
    EdgeParams(
        source_node="input_data",
        target_node="step_1",
        output_slot="all",
        input_slot="step_1_main_input",
    ),
    EdgeParams(
        source_node="step_1",
        target_node="step_2",
        output_slot="step_1_main_output",
        input_slot="step_2_main_input",
    ),
    EdgeParams(
        source_node="step_2",
        target_node="step_3",
        output_slot="step_2_main_output",
        input_slot="step_3_main_input",
    ),
    EdgeParams(
        source_node="step_3",
        target_node="choice_section",
        output_slot="step_3_main_output",
        input_slot="choice_section_main_input",
    ),
    EdgeParams(
        source_node="choice_section",
        target_node="results",
        output_slot="choice_section_main_output",
        input_slot="result",
    ),
    EdgeParams(
        source_node="input_data",
        target_node="choice_section",
        output_slot="all",
        input_slot="choice_section_secondary_input",
    ),
]
SCHEMA_PARAMS = (NODES, EDGES)



"""
======
Runner
======

This module contains the main function for running a pipeline; it is intended to
be called from the ``easylink.cli`` module.

"""

import os
import socket
import subprocess
from pathlib import Path

from graphviz import Source
from loguru import logger
from snakemake.cli import main as snake_main

from easylink.configuration import Config, load_params_from_specification
from easylink.pipeline import Pipeline
from easylink.utilities.data_utils import (
    copy_configuration_files_to_results_directory,
    create_results_directory,
    create_results_intermediates,
)
from easylink.utilities.general_utils import is_on_slurm
from easylink.utilities.paths import EASYLINK_TEMP


def main(
    command: str,
    pipeline_specification: str,
    input_data: str,
    computing_environment: str | None,
    results_dir: str,
    debug=False,
) -> None:
    """Runs an EasyLink command.

    This function is used to run an EasyLink job and is intended to be accessed via
    the ``easylink.cli`` module's command line interface (CLI) commands. It
    configures the run and sets up the pipeline based on the user-provided specification
    files and then calls on `Snakemake <https://snakemake.readthedocs.io/en/stable/>`_
    to act as the workflow manager.

    Arguments
    ---------
    command
        The command to run. Current supported commands include "run" and "generate_dag".
    pipeline_specification
        The filepath to the pipeline specification file.
    input_data
        The filepath to the input data specification file (_not_ the paths to the
        input data themselves).
    computing_environment
        The filepath to the specification file defining the computing environment
        to run the pipeline on. If None, the pipeline will be run locally.
    results_dir
        The directory to write results and incidental files (logs, etc.) to.
    debug
        If False (the default), will suppress some of the workflow output. This
        is intended to only be used for testing and development purposes.
    """
    config_params = load_params_from_specification(
        pipeline_specification, input_data, computing_environment, results_dir
    )
    config = Config(config_params)
    pipeline = Pipeline(config)
    # After validation is completed, create the results directory
    create_results_directory(Path(results_dir))
    snakefile = pipeline.build_snakefile()
    _save_dag_image(snakefile, results_dir)
    if command == "generate_dag":
        return
    # Copy the configuration files to the results directory if we actually plan to run the pipeline.
    create_results_intermediates(Path(results_dir))
    copy_configuration_files_to_results_directory(
        Path(pipeline_specification),
        Path(input_data),
        Path(computing_environment) if computing_environment else computing_environment,
        Path(results_dir),
    )
    environment_args = _get_environment_args(config)
    singularity_args = _get_singularity_args(config)
    # Set source cache in appropriate location to avoid jenkins failures
    os.environ["XDG_CACHE_HOME"] = results_dir + "/.snakemake/source_cache"
    # We need to set a dummy environment variable to avoid logging a wall of text.
    # TODO [MIC-4920]: Remove when https://github.com/snakemake/snakemake-interface-executor-plugins/issues/55 merges
    os.environ["foo"] = "bar"
    argv = [
        "--snakefile",
        str(snakefile),
        "--directory",
        results_dir,
        "--cores",
        "all",
        "--jobs",
        "unlimited",
        "--latency-wait=120",
        ## See above
        "--envvars",
        "foo",
        "--use-singularity",
        "--singularity-args",
        singularity_args,
    ]
    if not debug:
        # Suppress some of the snakemake output
        argv += [
            "--quiet",
            "progress",
        ]
    argv.extend(environment_args)
    logger.info(f"Running Snakemake")
    logger.debug(f"Snakemake arguments: {argv}")
    snake_main(argv)


def _get_singularity_args(config: Config) -> str:
    """Gets the required singularity arguments."""
    input_file_paths = ",".join(
        file.as_posix() for file in config.input_data.to_dict().values()
    )
    singularity_args = "--no-home --containall"
    easylink_tmp_dir = EASYLINK_TEMP[config.computing_environment]
    easylink_tmp_dir.mkdir(parents=True, exist_ok=True)
    singularity_args += f" -B {easylink_tmp_dir}:/tmp,$(pwd),{input_file_paths} --pwd $(pwd)"
    return singularity_args


def _get_environment_args(config: Config) -> list[str]:
    """Gets the required environment arguments."""
    # Set up computing environment
    if config.computing_environment == "local":
        return []

        # TODO [MIC-4822]: launch a local spark cluster instead of relying on implementation
    elif config.computing_environment == "slurm":
        if not is_on_slurm():
            raise RuntimeError(
                f"A 'slurm' computing environment is specified but it has been "
                "determined that the current host is not on a slurm cluster "
                f"(host: {socket.gethostname()})."
            )
        resources = config.slurm_resources
        slurm_args = ["--executor", "slurm", "--default-resources"] + [
            f"{resource_key}={resource_value}"
            for resource_key, resource_value in resources.items()
        ]
        return slurm_args
    else:
        raise NotImplementedError(
            "only computing_environment 'local' and 'slurm' are supported; "
            f"provided {config.computing_environment}"
        )


def _save_dag_image(snakefile, results_dir) -> None:
    """Saves the directed acyclic graph (DAG) of the pipeline to an image file.

    Attributes
    ----------
    snakefile
        The path to the snakefile.
    results_dir
        The directory to save the DAG image to.
    """
    process = subprocess.run(
        ["snakemake", "--snakefile", str(snakefile), "--dag"],
        capture_output=True,
        text=True,
        check=True,
    )
    dot_output = process.stdout
    source = Source(dot_output)
    # Render the graph to a file
    source.render("DAG", directory=results_dir, format="svg", cleanup=True)



"""
========================
Data Splitting Utilities
========================

This module contains utility functions for splitting datasets into smaller datasets.
One primary use case for this is to run sections of the pipeline in an embarrassingly 
parallel manner.

Note that it is critical that all data splitting utility functions are definied
in this module; easylink will not be able to find them otherwise.
"""

import math
import os

import pandas as pd
from loguru import logger


def split_data_by_size(
    input_files: list[str], output_dir: str, desired_chunk_size_mb: int | float
) -> None:
    """Splits the data (from a single input slot) into chunks of desired size.

    This function takes all datasets from a single input slot, concatenates them,
    and then splits the resulting dataset into chunks of the desired size. Note
    that this will split the data as evenly as possible, but the final chunk may
    be smaller than the desired size if the input data does not divide evenly; it
    makes no effort to redistribute the lingering data.

    Parameters
    ----------
    input_files
        A list of input file paths to be concatenated and split.
    output_dir
        The directory where the resulting chunks will be saved.
    desired_chunk_size_mb
        The desired size of each chunk, in megabytes.
    """

    # concatenate all input files
    df = pd.DataFrame()
    input_file_size_mb = 0
    for file in input_files:
        input_file_size_mb += os.path.getsize(file) / 1024**2
        tmp = pd.read_parquet(file)
        df = pd.concat([df, tmp], ignore_index=True)

    # divide df into num_chunks and save each one out
    num_chunks = math.ceil(input_file_size_mb / desired_chunk_size_mb)
    chunk_size = math.ceil(len(df) / num_chunks)
    if num_chunks == 1:
        logger.info(f"Input data is already smaller than desired chunk size; not splitting")
    else:
        logger.info(
            f"Splitting a {round(input_file_size_mb, 2)} MB dataset ({len(df)} rows) into "
            f"into {num_chunks} chunks of size ~{desired_chunk_size_mb} MB each"
        )
    for i in range(num_chunks):
        start = i * chunk_size
        end = (i + 1) * chunk_size
        chunk = df.iloc[start:end]
        chunk_dir = os.path.join(output_dir, f"chunk_{i}")
        if not os.path.exists(chunk_dir):
            os.makedirs(chunk_dir)
        logger.debug(
            f"Writing out chunk {i+1}/{num_chunks} (rows {chunk.index[0]} to "
            f"{chunk.index[-1]})"
        )
        chunk.to_parquet(os.path.join(chunk_dir, "result.parquet"))



# mypy: ignore-errors
import os
import re
import subprocess
import tempfile

import pytest
import yaml

from easylink.pipeline_schema import PipelineSchema
from easylink.pipeline_schema_constants import TESTING_SCHEMA_PARAMS
from easylink.runner import main
from easylink.utilities.general_utils import is_on_slurm
from tests.conftest import RESULTS_DIR, SPECIFICATIONS_DIR


@pytest.mark.slow
@pytest.mark.skipif(
    not is_on_slurm(),
    reason="Must be on slurm to run this test.",
)
def test_spark_slurm(mocker, caplog):
    """Test that the pipeline runs spark on SLURM with appropriate resources."""
    nodes, edges = TESTING_SCHEMA_PARAMS["integration"]
    mocker.patch("easylink.pipeline_schema.ALLOWED_SCHEMA_PARAMS", TESTING_SCHEMA_PARAMS)
    mocker.patch(
        "easylink.configuration.Config._get_schema",
        return_value=PipelineSchema("integration", nodes=nodes, edges=edges),
    )
    results_dir = tempfile.mkdtemp(dir=RESULTS_DIR)
    # give the tmpdir the same permissions as the parent directory so that
    # cluster jobs can write to it
    os.chmod(results_dir, os.stat(RESULTS_DIR).st_mode)
    pipeline_specification = SPECIFICATIONS_DIR / "integration/pipeline_spark.yaml"
    input_data = SPECIFICATIONS_DIR / "common/input_data.yaml"
    computing_environment = SPECIFICATIONS_DIR / "integration/environment_spark_slurm.yaml"
    with open(computing_environment, "r") as stream:
        env_config = yaml.safe_load(stream)
    with pytest.raises(SystemExit) as exit:
        main(
            command="run",
            pipeline_specification=pipeline_specification,
            input_data=input_data,
            computing_environment=computing_environment,
            results_dir=str(results_dir),
            debug=True,
        )
    assert exit.value.code == 0
    output = caplog.text
    job_ids = re.findall(r"Job \d+ has been submitted with SLURM jobid (\d+)", output)
    assert len(job_ids) == 3
    for idx, job_id in enumerate(job_ids):
        cmd = [
            "sacct",
            f"--jobs={job_id}",
            "--format=JobID,Account,Partition,ReqMem,ReqCPUS,TimelimitRaw",
            "--noheader",
            "--parsable2",
        ]
        result = subprocess.run(cmd, capture_output=True, text=True)
        output = result.stdout
        # Filter out jobs that are not the main job and grab full line
        main_job_pattern = r"^(\d+)\|"
        main_line = None
        for line in output.split("\n"):
            if re.match(main_job_pattern, line):
                main_line = line
                break
        assert main_line
        fields = main_line.split("|")
        account, partition, mem, cpus, time = fields[1:6]
        assert account == env_config["slurm"]["account"]
        assert partition == env_config["slurm"]["partition"]
        # First two jobs are spark master and spark worker
        if idx in [0, 1]:
            assert (
                mem
                == str(int(env_config["spark"]["workers"]["mem_per_node"] * 1024 + 500)) + "M"
            )
            assert cpus == str(env_config["spark"]["workers"]["cpus_per_node"])
            assert time == str(int(env_config["spark"]["workers"]["time_limit"] * 60))
        else:
            assert mem == str(env_config["implementation_resources"]["memory"]) + "G"
            assert cpus == str(env_config["implementation_resources"]["cpus"])
            assert time == str(int(env_config["implementation_resources"]["time_limit"] * 60))



# mypy: ignore-errors
import errno

import pytest

from easylink.utilities.general_utils import exit_with_validation_error


def test_exit_with_validation_error():
    with pytest.raises(SystemExit) as e:
        exit_with_validation_error({"error": "error message"})
    assert e.value.code == errno.EINVAL



import glob
import logging
import os

import pandas as pd
import yaml

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(message)s",
    handlers=[logging.StreamHandler()],
)


def load_file(file_path, file_format=None):
    if file_format is None:
        file_format = file_path.split(".")[-1]
    if file_format == "parquet":
        return pd.read_parquet(file_path)
    if file_format == "csv":
        return pd.read_csv(file_path)
    raise ValueError()


INPUT_ENV_VARS = os.getenv(
    "INPUT_ENV_VARS",
    "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
).split(",")

diagnostics = {}
df = pd.DataFrame()

for env_var in INPUT_ENV_VARS:
    if env_var not in os.environ:
        logging.error(f"Missing required environment variable {env_var}")
        raise ValueError()

    logging.info(f"Loading files for {env_var}")
    file_paths = os.environ[env_var].split(",")
    diagnostics[f"num_files_{env_var.lower()}"] = len(file_paths)
    for path in file_paths:
        df = pd.concat([df, load_file(path)], ignore_index=True).fillna(0)


extra_implementation_specific_input_glob = glob.glob(
    "/extra_implementation_specific_input_data/input*"
)
extra_implementation_specific_input_file_path = os.getenv(
    "DUMMY_CONTAINER_EXTRA_IMPLEMENTATION_SPECIFIC_INPUT_FILE_PATH",
    (
        extra_implementation_specific_input_glob[0]
        if len(extra_implementation_specific_input_glob) > 0
        else None
    ),
)
diagnostics["extra_implementation_specific_input"] = (
    extra_implementation_specific_input_file_path is not None
)
if extra_implementation_specific_input_file_path is not None:
    logging.info("Loading extra, implementation-specific input")
    df = pd.concat(
        [df, load_file(extra_implementation_specific_input_file_path)], ignore_index=True
    ).fillna(0)

logging.info(f"Total input length is {len(df)}")

broken = os.getenv("DUMMY_CONTAINER_BROKEN", "false").lower() in ("true", "yes", "1")
diagnostics["broken"] = broken
if broken:
    df = df.rename(
        columns={
            "foo": "wrong",
            "bar": "column",
            "counter": "names",
        }
    )
else:
    increment = int(os.getenv("DUMMY_CONTAINER_INCREMENT", "1"))
    diagnostics["increment"] = increment
    logging.info(f"Increment is {increment}")
    df["counter"] = df.counter + increment

    added_columns_existing = [c for c in df.columns if "added_column_" in c]
    diagnostics["added_columns_existing"] = added_columns_existing
    added_columns_existing_ints = [int(c.split("_")[-1]) for c in added_columns_existing]
    max_added_column = max(added_columns_existing_ints, default=0) + increment
    min_added_column = max(max_added_column - 4, 0)
    added_columns_desired = range(min_added_column, max_added_column + 1)
    added_columns_desired_names = [f"added_column_{i}" for i in added_columns_desired]
    diagnostics["added_columns_desired_names"] = added_columns_desired_names
    diagnostics["new_columns"] = []
    for column_index, column_name in zip(added_columns_desired, added_columns_desired_names):
        if column_name not in df.columns:
            diagnostics["new_columns"].append(column_name)
            df[column_name] = column_index

    columns_to_drop = [
        c for c in df.columns if "added_column_" in c and c not in added_columns_desired_names
    ]
    diagnostics["columns_to_drop"] = columns_to_drop
    df.drop(columns=columns_to_drop, inplace=True)

output_file_format = os.getenv("DUMMY_CONTAINER_OUTPUT_FILE_FORMAT", "parquet")
output_file_paths = os.getenv(
    "DUMMY_CONTAINER_OUTPUT_PATHS", f"/results/result.{output_file_format}"
).split(",")

diagnostics["num_output_files"] = len(output_file_paths)
diagnostics["output_file_paths"] = output_file_paths

for output_file_path in output_file_paths:
    logging.info(f"Writing output to {output_file_path} in {output_file_format} format")
    if output_file_format == "parquet":
        df.to_parquet(output_file_path)
    elif output_file_format == "csv":
        df.to_csv(output_file_path, index=False)
    else:
        raise ValueError()

diagnostics_dir = os.getenv("DUMMY_CONTAINER_DIAGNOSTICS_DIRECTORY", "/diagnostics")
try:
    with open(f"{diagnostics_dir}/diagnostics.yaml", "w") as f:
        yaml.dump(diagnostics, f, default_flow_style=False)
except (PermissionError, OSError):
    pass



""" This module contains tests for the validation methods for the various
classes in this package. These pseudo-unit tests have been separated from the
class- and module-specific test modules simply to make them easier to find and
maintain.
"""

import errno
import re

import pytest

from easylink.configuration import (
    ENVIRONMENT_ERRORS_KEY,
    INPUT_DATA_ERRORS_KEY,
    PIPELINE_ERRORS_KEY,
    Config,
    load_params_from_specification,
)
from easylink.pipeline import IMPLEMENTATION_ERRORS_KEY, Pipeline
from easylink.utilities import paths
from easylink.utilities.data_utils import load_yaml


def _check_expected_validation_exit(error, caplog, error_no, expected_msg):
    """Check that the validation messages are as expected. It's hacky."""
    assert error.value.code == error_no
    # Extract error message
    msg = caplog.text.split("Validation errors found. Please see below.")[1].split(
        "Validation errors found. Please see above."
    )[0]
    msg = re.sub("\n+", " ", msg)
    msg = re.sub(" +", " ", msg).strip()
    # Remove single quotes from msg and expected b/c they're difficult to handle and not that important
    msg = re.sub("'+", "", msg)

    def append_to_pattern(pattern, context):
        for item, messages in context.items():
            pattern.append(" " + item + ":")
            if isinstance(messages, dict):
                append_to_pattern(pattern, messages)
            else:
                for message in messages:
                    message = re.sub("'+", "", message)
                    pattern.append(" - " + message)
        return pattern

    all_matches = []
    for error_type, context in expected_msg.items():
        expected_pattern = append_to_pattern([error_type + ":"], context)
        pattern = re.compile("".join(expected_pattern))
        match = pattern.search(msg)
        assert match
        all_matches.append(match)

    covered_text = "".join(match.group(0) for match in all_matches)
    assert len(covered_text) == len(msg)


@pytest.mark.skip(reason="TODO [MIC-4735]")
def test_batch_validation():
    pass


######################
# Config validations #
######################


@pytest.mark.parametrize(
    "pipeline, expected_msg",
    [
        (
            "missing_implementations",
            {
                PIPELINE_ERRORS_KEY: {
                    "development": {
                        "step step_1": [
                            "The step configuration does not contain an 'implementation' key or a reference to a combined implementation."
                        ]
                    },
                },
            },
        ),
        (
            "missing_implementation_name",
            {
                PIPELINE_ERRORS_KEY: {
                    "development": {
                        "step step_1": [
                            "The implementation configuration does not contain a 'name' key."
                        ]
                    },
                },
            },
        ),
        (
            "missing_step",
            {
                PIPELINE_ERRORS_KEY: {
                    "development": {
                        "step choice_section": ["The step is not configured."],
                        "step step_1": ["The step is not configured."],
                        "step step_3": ["The step is not configured."],
                    },
                },
            },
        ),
        (
            "missing_loop_nodes",
            {
                PIPELINE_ERRORS_KEY: {
                    "development": {
                        "step step_3": ["No loop instances configured under iterate key."],
                    },
                },
            },
        ),
        (
            "bad_loop_formatting",
            {
                PIPELINE_ERRORS_KEY: {
                    "development": {
                        "step step_3": [
                            "Loop instances must be formatted as a sequence in the pipeline configuration."
                        ],
                    },
                },
            },
        ),
        (
            "missing_substeps",
            {
                PIPELINE_ERRORS_KEY: {
                    "development": {
                        "step step_4a": [
                            "The implementation configuration does not contain a 'name' key."
                        ],
                        "step step_4b": ["The step is not configured."],
                    },
                },
            },
        ),
        (
            "wrong_parallel_split_keys",
            {
                PIPELINE_ERRORS_KEY: {
                    "development": {
                        "step step_1": {
                            "parallel_split_1": {
                                "Input Data Key": [
                                    "Input data file 'foo' not found in input data configuration."
                                ],
                            },
                        },
                    },
                },
            },
        ),
        (
            "bad_combined_implementations",
            {
                PIPELINE_ERRORS_KEY: {
                    "development": {
                        "step step_3": [
                            "The step refers to a combined implementation but 'foo' is not a valid combined implementation."
                        ]
                    }
                }
            },
        ),
        (
            "missing_type_key",
            {
                PIPELINE_ERRORS_KEY: {
                    "development": {
                        "step choice_section": ["The step requires a 'type' key."]
                    }
                }
            },
        ),
        (
            "bad_type_key",
            {
                PIPELINE_ERRORS_KEY: {
                    "development": {
                        "step choice_section": [
                            "'foo' is not a supported 'type'. Valid choices are: \['simple', 'complex'\]."
                        ]
                    }
                }
            },
        ),
        (
            "type_config_mismatch",
            {
                PIPELINE_ERRORS_KEY: {
                    "development": {
                        "step choice_section": [
                            "Invalid configuration for 'simple' type. Valid steps are \['step_4'\]."
                        ]
                    }
                }
            },
        ),
    ],
)
def test_pipeline_validation(
    pipeline, expected_msg, default_config_params, unit_test_specifications_dir, caplog
):
    config_params = default_config_params
    config_params["pipeline"] = load_yaml(
        f"{unit_test_specifications_dir}/pipeline_{pipeline}.yaml"
    )

    with pytest.raises(SystemExit) as e:
        Config(config_params)

    _check_expected_validation_exit(
        error=e,
        caplog=caplog,
        error_no=errno.EINVAL,
        expected_msg=expected_msg,
    )


def test_out_of_order_steps(default_config_params, unit_test_specifications_dir):
    # Make sure we DON'T raise an exception even if the steps are out of order
    config_params = default_config_params
    config_params["pipeline"] = load_yaml(
        f"{unit_test_specifications_dir}/pipeline_out_of_order.yaml"
    )
    Config(config_params)


def test_unsupported_step(
    default_config_params, unit_test_specifications_dir, caplog, mocker
):
    mocker.patch("easylink.implementation.Implementation._load_metadata")
    mocker.patch("easylink.implementation.Implementation.validate", return_value=[])
    config_params = default_config_params
    config_params["pipeline"] = load_yaml(
        f"{unit_test_specifications_dir}/pipeline_bad_step.yaml"
    )

    with pytest.raises(SystemExit) as e:
        Config(config_params)

    _check_expected_validation_exit(
        error=e,
        caplog=caplog,
        error_no=errno.EINVAL,
        expected_msg={
            PIPELINE_ERRORS_KEY: {
                "development": {
                    "step foo": ["foo is not a valid step."],
                }
            }
        },
    )


def test_unsupported_implementation(
    default_config_params, unit_test_specifications_dir, caplog, mocker
):
    mocker.patch("easylink.implementation.Implementation._load_metadata")
    mocker.patch("easylink.implementation.Implementation.validate", return_value=[])
    config_params = default_config_params
    config_params["pipeline"] = load_yaml(
        f"{unit_test_specifications_dir}/pipeline_bad_implementation.yaml"
    )

    with pytest.raises(SystemExit) as e:
        Config(config_params)

    implementation_metadata = load_yaml(paths.IMPLEMENTATION_METADATA)
    supported_implementations = (
        str(list(implementation_metadata.keys())).replace("[", "\\[").replace("]", "\\]")
    )
    _check_expected_validation_exit(
        error=e,
        caplog=caplog,
        error_no=errno.EINVAL,
        expected_msg={
            PIPELINE_ERRORS_KEY: {
                "development": {
                    "step step_1": [
                        f"Implementation 'foo' is not supported. Supported implementations are: {supported_implementations}."
                    ],
                },
            }
        },
    )


def test_pipeline_schema_bad_input_data_type(default_config_paths, test_dir, caplog):
    config_paths = default_config_paths
    config_params = load_params_from_specification(**config_paths)
    config_params["input_data"] = {}
    with pytest.raises(SystemExit) as e:
        Config(config_params)

    _check_expected_validation_exit(
        error=e,
        caplog=caplog,
        error_no=errno.EINVAL,
        expected_msg={INPUT_DATA_ERRORS_KEY: ["No input data is configured."]},
    )


def test_pipeline_schema_bad_input_data_type(default_config_paths, test_dir, caplog):
    config_paths = default_config_paths
    config_paths.update(
        {"input_data": f"{test_dir}/bad_type_input_data.yaml", "computing_environment": None}
    )
    config_params = load_params_from_specification(**config_paths)
    with pytest.raises(SystemExit) as e:
        Config(config_params)

    _check_expected_validation_exit(
        error=e,
        caplog=caplog,
        error_no=errno.EINVAL,
        expected_msg={
            INPUT_DATA_ERRORS_KEY: {
                r".*/file1.oops": [r"Data file type .oops is not supported. Convert to .*"],
            },
        },
    )


def test_pipeline_schema_bad_input_data(default_config_paths, test_dir, caplog):
    config_paths = default_config_paths
    config_paths.update(
        {
            "input_data": f"{test_dir}/bad_columns_input_data.yaml",
            "computing_environment": None,
        }
    )
    config_params = load_params_from_specification(**config_paths)
    with pytest.raises(SystemExit) as e:
        Config(config_params)

    _check_expected_validation_exit(
        error=e,
        caplog=caplog,
        error_no=errno.EINVAL,
        expected_msg={
            INPUT_DATA_ERRORS_KEY: {
                ".*/broken_file1.csv": [r"Data file .* is missing required column\(s\) .*"],
            }
        },
    )


def test_pipeline_schema_missing_input_file(default_config_paths, test_dir, caplog):
    config_paths = default_config_paths
    config_paths.update(
        {"input_data": f"{test_dir}/missing_input_data.yaml", "computing_environment": None}
    )
    config_params = load_params_from_specification(**config_paths)
    with pytest.raises(SystemExit) as e:
        Config(config_params)

    _check_expected_validation_exit(
        error=e,
        caplog=caplog,
        error_no=errno.EINVAL,
        expected_msg={
            INPUT_DATA_ERRORS_KEY: {
                ".*/missing_file1.csv": ["File not found."],
            },
        },
    )


# Environment validations
def test_unsupported_container_engine(default_config_params, caplog):
    config_params = default_config_params
    config_params["environment"] = {"container_engine": "foo"}
    with pytest.raises(SystemExit) as e:
        Config(config_params)
    _check_expected_validation_exit(
        error=e,
        caplog=caplog,
        error_no=errno.EINVAL,
        expected_msg={
            "ENVIRONMENT ERRORS": {
                "container_engine": ["The value 'foo' is not supported."],
            },
        },
    )


def test_missing_slurm_details(default_config_params, caplog):
    config_params = default_config_params
    config_params["environment"] = {"computing_environment": "slurm"}
    with pytest.raises(SystemExit) as e:
        Config(config_params)
    _check_expected_validation_exit(
        error=e,
        caplog=caplog,
        error_no=errno.EINVAL,
        expected_msg={
            ENVIRONMENT_ERRORS_KEY: {
                "slurm": [
                    "The environment configuration file must include a 'slurm' key "
                    "defining slurm resources if the computing_environment is 'slurm'."
                ],
            },
        },
    )


############
# pipeline #
############


def test_no_container(default_config, caplog, mocker):
    metadata = load_yaml(paths.IMPLEMENTATION_METADATA)
    metadata["step_1_python_pandas"]["image_path"] = "some/path/with/no/container.sif"
    metadata["step_2_python_pandas"]["image_path"] = "some/path/with/no/container_2.sif"
    metadata["step_3_python_pandas"]["image_path"] = "some/path/with/no/container_3.sif"
    metadata["step_4_python_pandas"]["image_path"] = "some/path/with/no/container_4.sif"
    mocker.patch("easylink.implementation.load_yaml", return_value=metadata)
    mocker.PropertyMock(
        "easylink.implementation.Implementation._container_engine", return_value="undefined"
    )
    with pytest.raises(SystemExit) as e:
        Pipeline(default_config)

    _check_expected_validation_exit(
        error=e,
        caplog=caplog,
        error_no=errno.EINVAL,
        expected_msg={
            IMPLEMENTATION_ERRORS_KEY: {
                "step_1_python_pandas": [
                    "Container 'some/path/with/no/container.sif' does not exist.",
                ],
                "step_2_python_pandas": [
                    "Container 'some/path/with/no/container_2.sif' does not exist.",
                ],
                "step_3_python_pandas": [
                    "Container 'some/path/with/no/container_3.sif' does not exist.",
                ],
                "step_4_python_pandas": [
                    "Container 'some/path/with/no/container_4.sif' does not exist.",
                ],
            },
        },
    )


def test_implemenation_does_not_match_step(default_config, caplog, mocker):
    metadata = load_yaml(paths.IMPLEMENTATION_METADATA)
    metadata["step_1_python_pandas"]["steps"] = ["not-the-step-1-name"]
    metadata["step_2_python_pandas"]["steps"] = ["not-the-step-2-name"]
    mocker.patch("easylink.implementation.load_yaml", return_value=metadata)
    mocker.patch(
        "easylink.implementation.Implementation._validate_container_exists",
        side_effect=lambda x: x,
    )

    with pytest.raises(SystemExit) as e:
        Pipeline(default_config)

    _check_expected_validation_exit(
        error=e,
        caplog=caplog,
        error_no=errno.EINVAL,
        expected_msg={
            IMPLEMENTATION_ERRORS_KEY: {
                "step_1_python_pandas": [
                    r"Pipeline configuration nodes \['step_1'\] do not match metadata steps \['not-the-step-1-name'\]."
                ],
                "step_2_python_pandas": [
                    r"Pipeline configuration nodes \['step_2'\] do not match metadata steps \['not-the-step-2-name'\]."
                ],
            },
        },
    )



"""
===============
Snakemake Rules
===============

We have chosen to use `Snakemake <https://snakemake.readthedocs.io/en/stable/>`_ 
as the EasyLink workflow manager. This module is responsible for generating the 
Snakemake rules to be run as well as writing them to the Snakefile.

Note we have adopted the Snakemake term "rule" to refer to a singular component 
in a Snakefile (i.e. in a Snakemake pipeline) that defines input files, output files,
and the command to run to create those output files. These rules are generated
dynamically as strings and appended to the Snakefile.
"""

import os
from abc import ABC, abstractmethod
from collections.abc import Callable
from dataclasses import dataclass


class Rule(ABC):
    """An abstract class used to generate Snakemake rules."""

    def write_to_snakefile(self, snakefile_path) -> None:
        """Writes the rule to the Snakefile.

        Parameters
        ----------
        snakefile_path
            Path to the Snakefile to write the rule to.
        """
        with open(snakefile_path, "a") as f:
            f.write(self.build_rule())

    @abstractmethod
    def build_rule(self) -> str:
        """Builds the snakemake rule to be written to the Snakefile.

        This is an abstract method and must be implemented by concrete instances.
        """
        pass

    @staticmethod
    def get_input_slots_to_split(input_slots) -> list[str]:
        input_slots_to_split = [
            slot_name
            for slot_name, slot_attrs in input_slots.items()
            if slot_attrs.get("splitter", None)
        ]
        return input_slots_to_split


@dataclass
class TargetRule(Rule):
    """A :class:`Rule` that defines the final output of the pipeline.

    Snakemake will determine the directed acyclic graph (DAG) based on this target.
    """

    target_files: list[str]
    """List of final output filepaths."""
    validation: str
    """Name of file created by :class:`InputValidationRule`."""
    requires_spark: bool
    """Whether or not this rule requires a Spark environment to run."""

    def build_rule(self) -> str:
        """Builds the Snakemake rule for the final output of the pipeline."""
        outputs = [os.path.basename(file_path) for file_path in self.target_files]
        rulestring = f"""
rule all:
    message: 'Grabbing final output'
    localrule: True   
    input:
        final_output={self.target_files},
        validation='{self.validation}',"""
        if self.requires_spark:
            rulestring += f"""
        term="spark_logs/spark_master_terminated.txt",
        master_log="spark_logs/spark_master_log.txt",
        worker_logs=gather.num_workers("spark_logs/spark_worker_log_{{scatteritem}}.txt",
        ),"""
        rulestring += f"""
    output: {outputs}
    run:
        import os
        for input_path, output_path in zip(input.final_output, output):
            os.symlink(input_path, output_path)"""
        return rulestring


@dataclass
class ImplementedRule(Rule):
    """A :class:`Rule` that defines the execution of an :class:`~easylink.implementation.Implementation`."""

    name: str
    """Name of the rule."""
    step_name: str
    """Name of the step this rule is implementing."""
    implementation_name: str
    """Name of the ``Implementation`` to build the rule for."""
    input_slots: dict[str, dict[str, str | list[str]]]
    """This ``Implementation's`` input slot attributes."""
    validations: list[str]
    """Names of files created by :class:`InputValidationRule`. These files are empty
    but used by Snakemake to build the graph edges of dependency on validation rules."""
    output: list[str]
    """Output data filepaths."""
    resources: dict | None
    """Computational resources used by executor (e.g. SLURM)."""
    envvars: dict
    """Environment variables to set."""
    diagnostics_dir: str
    """Directory for diagnostic files."""
    image_path: str
    """Path to the Singularity image to run."""
    script_cmd: str
    """Command to execute."""
    requires_spark: bool
    """Whether or not this ``Implementation`` requires a Spark environment."""
    is_embarrassingly_parallel: bool = False
    """Whether or not this ``Implementation`` is to be run in an embarrassingly 
    parallel way."""

    def build_rule(self) -> str:
        """Builds the Snakemake rule for this ``Implementation``."""
        return self._build_io() + self._build_resources() + self._build_shell_cmd()

    def _build_io(self) -> str:
        """Builds the input/output portion of the rule."""
        if self.is_embarrassingly_parallel:
            # Processed chunks are sent to a 'processed' subdir
            output_files = [
                os.path.dirname(file_path)
                + "/processed/{chunk}/"
                + os.path.basename(file_path)
                for file_path in self.output
            ]
            log_path_chunk_adder = "-{chunk}"
        else:
            output_files = self.output
            log_path_chunk_adder = ""

        io_str = (
            f"""
rule:
    name: "{self.name}"
    message: "Running {self.step_name} implementation: {self.implementation_name}" """
            + self._build_input()
            + f"""        
    output: {output_files}
    log: "{self.diagnostics_dir}/{self.name}-output{log_path_chunk_adder}.log"
    container: "{self.image_path}" """
        )
        return io_str

    def _build_input(self) -> str:
        input_str = f"""
    input:"""
        input_slots_to_split = self.get_input_slots_to_split(self.input_slots)
        for slot, attrs in self.input_slots.items():
            env_var = attrs["env_var"].lower()
            if len(input_slots_to_split) > 1:
                raise NotImplementedError(
                    "FIXME [MIC-5883] Multiple input slots to split not yet supported"
                )
            if self.is_embarrassingly_parallel and slot == input_slots_to_split[0]:
                # The input to this is the input_chunks subdir from the checkpoint
                # rule (which is built by modifying the output of the overall implementation)
                if len(self.output) > 1:
                    raise NotImplementedError(
                        "FIXME [MIC-5883] Multiple output slots/files of EmbarrassinglyParallelSteps not yet supported"
                    )
                input_files = [
                    os.path.dirname(self.output[0])
                    + "/input_chunks/{chunk}/"
                    + os.path.basename(self.output[0])
                ]
            else:
                input_files = attrs["filepaths"]
            input_str += f"""
        {env_var}={input_files},"""
        if not self.is_embarrassingly_parallel:
            # validations were already handled in the checkpoint rule - no need
            # to validate the individual chunks
            input_str += f"""
        validations={self.validations},"""
        if self.requires_spark:
            input_str += f"""
        master_trigger=gather.num_workers(rules.wait_for_spark_worker.output),
        master_url=rules.wait_for_spark_master.output,"""
        return input_str

    def _build_resources(self) -> str:
        """Builds the resources portion of the rule."""
        if not self.resources:
            return ""
        return f"""
    resources:
        slurm_partition={self.resources['slurm_partition']},
        mem_mb={self.resources['mem_mb']},
        runtime={self.resources['runtime']},
        cpus_per_task={self.resources['cpus_per_task']},
        slurm_extra="--output '{self.diagnostics_dir}/{self.name}-slurm-%j.log'" """

    def _build_shell_cmd(self) -> str:
        """Builds the shell command portion of the rule."""
        # TODO [MIC-5787]: handle multiple wildcards, e.g.
        #   output_paths = ",".join(self.output)
        #   wildcards_subdir = "/".join([f"{{wildcards.{wc}}}" for wc in self.wildcards])
        #   and then in shell cmd: export DUMMY_CONTAINER_OUTPUT_PATHS={output_paths}/{wildcards_subdir}
        if self.is_embarrassingly_parallel:
            if len(self.output) > 1:
                raise NotImplementedError(
                    "FIXME [MIC-5883] Multiple output slots/files of EmbarrassinglyParallelSteps not yet supported"
                )
            output_files = (
                os.path.dirname(self.output[0])
                + "/processed/{wildcards.chunk}/"
                + os.path.basename(self.output[0])
            )
        else:
            output_files = ",".join(self.output)
        shell_cmd = f"""
    shell:
        '''
        export DUMMY_CONTAINER_OUTPUT_PATHS={output_files}
        export DUMMY_CONTAINER_DIAGNOSTICS_DIRECTORY={self.diagnostics_dir}"""
        for input_slot_name, input_slot_attrs in self.input_slots.items():
            input_slots_to_split = self.get_input_slots_to_split(self.input_slots)
            if len(input_slots_to_split) > 1:
                raise NotImplementedError(
                    "FIXME [MIC-5883] Multiple input slots to split not yet supported"
                )
            if input_slot_name in input_slots_to_split:
                # The inputs to this come from the input_chunks subdir
                input_files = (
                    os.path.dirname(self.output[0])
                    + "/input_chunks/{wildcards.chunk}/"
                    + os.path.basename(self.output[0])
                )
            else:
                input_files = ",".join(input_slot_attrs["filepaths"])
            shell_cmd += f"""
        export {input_slot_attrs["env_var"]}={input_files}"""
        if self.requires_spark:
            shell_cmd += f"""
        read -r DUMMY_CONTAINER_SPARK_MASTER_URL < {{input.master_url}}
        export DUMMY_CONTAINER_SPARK_MASTER_URL"""
        for var_name, var_value in self.envvars.items():
            shell_cmd += f"""
        export {var_name}={var_value}"""
        # Log stdout/stderr to diagnostics directory
        shell_cmd += f"""
        {self.script_cmd} > {{log}} 2>&1
        '''"""

        return shell_cmd


@dataclass
class InputValidationRule(Rule):
    """A :class:`Rule` that validates input files.

    Each file coming into the pipeline via an :class:`~easylink.graph_components.InputSlot`
    must be validated against a specific validator function. This rule is responsible
    for running those validations as well as creating the (empty) validation output
    files that are used by Snakemake to build the graph edge from this rule to the
    next.
    """

    name: str
    """Name of the rule."""
    input_slot_name: str
    """Name of the ``InputSlot``."""
    input: list[str]
    """List of filepaths to validate."""
    output: str
    """Filepath of validation output. It must be used as an input for next rule."""
    validator: Callable
    """Callable that takes a filepath as input. Raises an error if invalid."""

    def build_rule(self) -> str:
        """Builds the Snakemake rule for this validation.

        This rule runs the appropriate validator function on each input file as well
        as creates an empty file at the end. This empty file is used by Snakemake
        to build the graph edge from this rule to the next (since the validations
        themselves don't generate any output).
        """
        return f"""
rule:
    name: "{self.name}_{self.input_slot_name}_validator"
    input: {self.input}
    output: touch("{self.output}")
    localrule: True         
    message: "Validating {self.name} input slot {self.input_slot_name}"
    run:
        for f in input:
            validation_utils.{self.validator.__name__}(f)"""


@dataclass
class CheckpointRule(Rule):
    """A :class:`Rule` that defines a checkpoint.

    When running an :class:`~easylink.implementation.Implementation` in an embarrassingly
    parallel way, we do not know until runtime how many parallel jobs there will
    be (e.g. we don't know beforehand how many chunks a large incoming dataset will
    be split into since the incoming dataset isn't created until runtime). The
    snakemake mechanism to handle this dynamic nature is a
    `checkpoint <https://snakemake.readthedocs.io/en/stable/snakefiles/rules.html#data-dependent-conditional-execution/>`_
    rule along with a directory as output.

    Notes
    -----
    There is a known `Snakemake bug <https://github.com/snakemake/snakemake/issues/3036>`_
    which prevents the use of multiple checkpoints in a single Snakefile. We
    work around this by generating an empty checkpoint.txt file as part of this
    rule. If this file does not yet exist when trying to run the :class:`AggregationRule`,
    it means that the checkpoint has not yet been executed for the
    particular wildcard value(s). In this case, we manually raise a Snakemake
    ``IncompleteCheckpointException`` which Snakemake automatically handles
    and leads to a re-evaluation after the checkpoint has successfully passed.

    TODO [MIC-5658]: Thoroughly test this workaround when implementing cacheing.
    """

    name: str
    """Name of the rule."""
    input_slots: dict[str, dict[str, str | list[str]]]
    """This ``Implementation's`` input slot attributes."""
    validations: list[str]
    """Validation files from previous rule."""
    output: list[str]
    """Output directory path. It must be used as an input for next rule."""

    def build_rule(self) -> str:
        """Builds the Snakemake rule for this checkpoint.

        Checkpoint rules are a special type of rule in Snakemake that allow for dynamic
        generation of output files. This rule is responsible for splitting the input
        files into chunks. Note that the output of this rule is a Snakemake ``directory``
        object as opposed to a specific file like typical rules have.
        """
        # Replace the output filepath with an input_chunks subdir
        output_dir = os.path.dirname(self.output[0]) + "/input_chunks"
        input_slots_to_split = self.get_input_slots_to_split(self.input_slots)
        if len(input_slots_to_split) > 1:
            raise NotImplementedError(
                "FIXME [MIC-5883] Multiple input slots to split not yet supported"
            )
        input_slot_to_split = input_slots_to_split[0]
        checkpoint = f"""
checkpoint:
    name: "split_{self.name}_{input_slot_to_split}"
    input: 
        files={self.input_slots[input_slot_to_split]['filepaths']},
        validations={self.validations},
    output: 
        output_dir=directory("{output_dir}"),
        checkpoint_file=touch("{output_dir}/checkpoint.txt"),
    params:
        input_files=lambda wildcards, input: ",".join(input.files),
    localrule: True
    message: "Splitting {self.name} {input_slot_to_split} into chunks"
    run:
        splitter_utils.{self.input_slots[input_slot_to_split]["splitter"].__name__}(
            input_files=list(input.files),
            output_dir=output.output_dir,
            desired_chunk_size_mb=0.1,
        )"""
        return checkpoint


@dataclass
class AggregationRule(Rule):
    """A :class:`Rule` that aggregates the processed chunks of output data.

    When running an :class:`~easylink.implementation.Implementation` in an embarrassingly
    parallel way, we need to aggregate the output files from each parallel job
    into a single output file.
    """

    name: str
    """Name of the rule."""
    input_slots: dict[str, dict[str, str | list[str]]]
    """This ``Implementation's`` input slot attributes."""
    output_slot_name: str
    """Name of the :class:`~easylink.graph_components.OutputSlot`."""
    output_slot: dict[str, str | list[str]]
    """The output slot attributes to create this rule for."""

    def build_rule(self) -> str:
        """Builds the Snakemake rule for this aggregator.

        When running an :class:`~easylink.step.EmbarrassinglyParallelStep`, we need
        to aggregate the output files from each parallel job into a single output file.
        This rule relies on a dynamically generated aggregation function which returns
        all of the **processed** chunks (from running the ``EmbarrassinglyParallelStep's``
        container in parallel) and uses them as inputs to the actual aggregation
        rule.

        Notes
        -----
        There is a known `Snakemake bug <https://github.com/snakemake/snakemake/issues/3036>`_
        which prevents the use of multiple checkpoints in a single Snakefile. We
        work around this by generating an empty checkpoint.txt file in the
        :class:`~CheckpointRule`. If this file does not yet exist when trying to
        aggregate, it means that the checkpoint has not yet been executed for the
        particular wildcard value(s). In this case, we manually raise a Snakemake
        ``IncompleteCheckpointException`` which `Snakemake automatically handles
        <https://snakemake.readthedocs.io/en/stable/snakefiles/rules.html#data-dependent-conditional-execution>`_
        and leads to a re-evaluation after the checkpoint has successfully passed,
        i.e. we replicate `Snakemake's behavior <https://github.com/snakemake/snakemake/blob/04f89d330dd94baa51f41bc796392f85bccbd231/snakemake/checkpoints.py#L42>`_.
        """
        input_function = self._define_input_function()
        rule = self._define_aggregator_rule()
        return input_function + rule

    def _define_input_function(self):
        """Builds the `input function <https://snakemake.readthedocs.io/en/stable/snakefiles/rules.html#input-functions>`_."""
        if len(self.output_slot["filepaths"]) > 1:
            raise NotImplementedError(
                "FIXME [MIC-5883] Multiple output slots/files of EmbarrassinglyParallelSteps not yet supported"
            )
        if len(self.output_slot["filepaths"]) > 1:
            raise NotImplementedError(
                "FIXME [MIC-5883] Multiple slots/files of EmbarrassinglyParallelSteps not yet supported"
            )
        output_filepath = self.output_slot["filepaths"][0]
        checkpoint_file_path = (
            os.path.dirname(output_filepath) + "/input_chunks/checkpoint.txt"
        )
        input_slots_to_split = self.get_input_slots_to_split(self.input_slots)
        if len(input_slots_to_split) > 1:
            raise NotImplementedError(
                "FIXME [MIC-5883] Multiple input slots to split not yet supported"
            )
        input_slot_to_split = input_slots_to_split[0]
        checkpoint_name = f"checkpoints.split_{self.name}_{input_slot_to_split}"
        output_files = (
            os.path.dirname(output_filepath)
            + "/processed/{chunk}/"
            + os.path.basename(output_filepath)
        )
        func = f"""
def get_aggregation_inputs_{self.name}_{self.output_slot_name}(wildcards):
    checkpoint_file = "{checkpoint_file_path}"
    if not os.path.exists(checkpoint_file):
        output, _ = {checkpoint_name}.rule.expand_output(wildcards)
        raise IncompleteCheckpointException({checkpoint_name}.rule, checkpoint_target(output[0]))
    checkpoint_output = glob.glob(f"{{{checkpoint_name}.get(**wildcards).output.output_dir}}/*/")
    chunks = [Path(filepath).parts[-1] for filepath in checkpoint_output]
    return expand(
        "{output_files}",
        chunk=chunks
    )"""
        return func

    def _define_aggregator_rule(self):
        """Builds the rule that runs the aggregation."""
        rule = f"""
rule:
    name: "aggregate_{self.name}_{self.output_slot_name}"
    input: get_aggregation_inputs_{self.name}_{self.output_slot_name}
    output: {self.output_slot["filepaths"]}
    localrule: True
    message: "Aggregating {self.name} {self.output_slot_name}"
    run:
        aggregator_utils.{self.output_slot["aggregator"].__name__}(
            input_files=list(input),
            output_filepath="{self.output_slot["filepaths"][0]}",
        )"""
        return rule



# mypy: ignore-errors
import os

import pytest

from easylink.configuration import Config
from easylink.runner import _get_environment_args, _get_singularity_args
from easylink.utilities.data_utils import load_yaml
from easylink.utilities.paths import EASYLINK_TEMP

IN_GITHUB_ACTIONS = os.getenv("GITHUB_ACTIONS") == "true"


def test_get_singularity_args(default_config, test_dir):
    assert (
        _get_singularity_args(default_config)
        == f"--no-home --containall -B {EASYLINK_TEMP[default_config.computing_environment]}:/tmp,"
        f"$(pwd),"
        f"{test_dir}/input_data1/file1.csv,"
        f"{test_dir}/input_data2/file2.csv"
        " --pwd $(pwd)"
    )


def test_get_environment_args_local(default_config_params):
    config = Config(default_config_params)
    assert _get_environment_args(config) == []


@pytest.mark.skipif(
    IN_GITHUB_ACTIONS,
    reason="Github Actions does not have access to our file system and so no SLURM.",
)
def test_get_environment_args_slurm(default_config_params, unit_test_specifications_dir):
    slurm_config_params = default_config_params
    slurm_config_params["environment"] = load_yaml(
        f"{unit_test_specifications_dir}/environment_spark_slurm.yaml"
    )
    slurm_config = Config(slurm_config_params)
    resources = slurm_config.slurm_resources
    assert _get_environment_args(slurm_config) == [
        "--executor",
        "slurm",
        "--default-resources",
        f"slurm_account={resources['slurm_account']}",
        f"slurm_partition={resources['slurm_partition']}",
        f"mem_mb={resources['mem_mb']}",
        f"runtime={resources['runtime']}",
        f"cpus_per_task={resources['cpus_per_task']}",
    ]
